TEST_BW_PROJECT = "ecoinvent_391"
TEST_BW_DATABASE = "ecoinvent_391_cutoff"

data = {
    "bw_project": TEST_BW_PROJECT,
    "bw_default_database": TEST_BW_DATABASE,
    "activities": {
        "single_activity": {
            "id": {
                "name": "concentrated solar power plant construction, "
                "solar tower power plant, 20 MW",
                "code": "19978cf531d88e55aed33574e1087d78",
            },
            "output": ["unit", 1],
        }
    },
    "methods": [
        {"id": ["EDIP 2003 no LT", "non-renewable resources no LT", "zinc no LT"]}
    ],
}

# exp = Experiment(data)
# from enbios2.base.unit_registry import ureg

# from enbios2.base.unit_registry import ureg as o_ureg

# ureg.parse_expression("kWh")

# data_path: DataPath =
# DataPath("ecoinvent/ecoinvent 3.9.1_cutoff_ecoSpold02/MasterData/Units.xml")
# res = parse_xml(data_path, "unit", [], ["name", "comment"])
#
# for unit in res:
#     u = str(unit["name"])
#     try:
#         p = ureg.parse_expression(u)
#         print(u, " - ", p)
#     except Exception as e:
#         print(u, " ? ", unit["comment"])
#     finally:
#         print("******")

# print(ureg.parse_expression("kilowatt hour"))
# print(ureg.parse_expression("kilowatt hour") * 1)
# print(ureg.parse_expression("kilowatt_hour") * 1)
#
#
# print(o_ureg.parse_expression("kilowatt hour"))
#
# # ureg.define('kilowatt_hour = hour * kilowatt')
# print(o_ureg.parse_expression("kilowatt hour"))



from pathlib import Path

import csv
from typing import Union

from deprecated import deprecated


def build_tree_from_csv(csv_file: Path, level_cols: list[str], attr_maps: dict[str, str]) -> dict:
    # Placeholder for root of the tree
    root = None

    # list to store the last node at each level
    last_nodes = []

    with open(csv_file, newline='') as csvfile:
        reader = csv.reader(csvfile)

        # get header
        header = next(reader)

        # ensure all level columns are in the header
        for col in level_cols:
            if col not in header:
                raise ValueError(f"Level column {col} not found in CSV header")

        for row in reader:
            row_dict = dict(zip(header, row))

            for level, level_col in enumerate(level_cols):
                name = row_dict[level_col]

                if name != '':
                    # new node with attributes
                    new_node = {"name": name, "children": {}}
                    attr_map = attr_maps[level] if level < len(attr_maps) else {}
                    for col, attr in attr_map.items():
                        if row_dict[col] != '':
                            new_node[attr] = row_dict[col]

                    # if this is the root node
                    if level == 0:
                        root = new_node

                    # add as child of the last node at the previous level, if there is one
                    if level > 0:
                        last_nodes[level - 1]["children"][name] = new_node

                    # this node is now the last node at this level
                    if level >= len(last_nodes):
                        last_nodes.append(new_node)
                    else:
                        last_nodes[level] = new_node

    return root


import csv


@deprecated(reason="Use BasicTreeNode.to_csv")
def tree_to_csv(root: dict, csv_file: Path, *, include_attrs=None, level_names: list[str] = None,
                merge_first_sub_row: bool = False, repeat_parent_name: bool = False):
    # Calculate max_depth based on root if not provided
    if include_attrs is None:
        include_attrs = []
    _total_level_names = level_names if level_names else []

    def level_name(level: int) -> str:
        if level >= len(_total_level_names):
            _total_level_names.append(f"lvl{level}")
        return _total_level_names[level]

    def rec_add_node_row(node: dict, current_level: int = 0) -> list[dict[str, Union[str, float]]]:
        row = {}
        for attr in include_attrs:
            row[attr] = node.get(attr, "")
        row[level_name(current_level)] = node["name"]
        _sub_rows = []
        for child in node.get("children", []):
            _sub_rows.extend(rec_add_node_row(child, current_level + 1))
        if _sub_rows:
            if merge_first_sub_row:
                row = {**row, **_sub_rows[0]}
                _sub_rows = _sub_rows[1:]
            if repeat_parent_name:
                for sub_row in _sub_rows:
                    sub_row[level_name(current_level)] = node["name"]
        return [row] + _sub_rows

    # Write rows to csv
    with csv_file.open('w', newline='') as csvfile:
        rows = rec_add_node_row(root)
        headers = _total_level_names + include_attrs
        writer = csv.DictWriter(csvfile, headers)
        writer.writeheader()
        writer.writerows(rows)

# json_file = BASE_DATA_PATH / "temp/miquel_upscaling/complete.json"
# print(json_file.exists())
# root = json.load(json_file.open())["impacts"]
# tree_to_csv({
#     "name": "root",
#     "children": {
#         "a": {
#             "name": "a",
#             "children": {
#                 "a1": {
#                     "name": "a1",
#                     "children": {}
#                 },
#                 "a2": {
#                     "name": "a2",
#                     "children": {}
#                 }
#             }
#         },
#         "b": {
#             "name": "b",
#             "children": {}}}},
#     BASE_DATA_PATH / "temp/miquel_upscaling/test.csv", ["value"], repeat_parent_name=False,
#     merge_first_sub_row=True)



"""
Mostly deprecated, because we use the indexers now.
"""

import importlib
import inspect
from pathlib import Path
from typing import Any, TypeVar, Union

import bw2data as bd
import bw2io as bi

from enbios2.const import BASE_DATA_PATH
from enbios2.generic.enbios2_logging import get_logger
from enbios2.models.bw_project_models import BWProject, BWProjectDatabase

logger = get_logger(__file__)


def list_importers() -> dict[str, Union[dict[str, list[TypeVar]], dict[str, TypeVar]]]:
    """
    :return: format map to list of importers, names to importers map
    """
    package = importlib.import_module("bw2io.importers")
    format2importer: dict[str, list[TypeVar]] = {}
    name2importer: dict[str, TypeVar] = {}
    for name, obj in inspect.getmembers(package):
        if inspect.isclass(obj):
            format_value = getattr(obj, "format", None)
            if format_value is not None:
                format2importer.setdefault(format_value, []).append(obj)
                name2importer[name] = obj
                # print(format_value, name)
    return {"formats": format2importer, "importers": name2importer}


def get_bw_importer(bw_database: BWProjectDatabase) -> Any:
    if bw_database.importer:
        return list_importers()["importers"][bw_database.importer]
    else:
        formats = list_importers()["formats"][bw_database.format]
        if len(formats) == 1:
            return formats[0]
        else:
            raise Exception("Multiple importers for format, please specify importer")


def setup_bw_project(project: BWProject, require_fresh: bool = False) -> None:
    """
    Set up a Brightway2 project and import databases
    :param project:
    :param require_fresh:
    :return:
    """
    logger.info(f"Setup {project.project_name}")
    if project.project_name in bd.projects:
        if require_fresh:
            raise Exception(
                f"project {project.project_name} already exists and "
                f"'require_fresh' is set to True."
            )
        logger.info(f"project '{project.project_name}' already exists.")
        bd.projects.set_current(project.project_name)
    else:
        logger.info("creating project")
        bd.projects.create_project(project.project_name)
        bd.projects.set_current(project.project_name)
        bi.bw2setup()

    for db in project.databases:
        setup_bw_db(db)


def setup_bw_db(db: BWProjectDatabase):
    if db.name in bd.databases:
        logger.info(f"Database {db.name} already exists, skipping")
        return
    path = Path(db.source)
    if not path.is_absolute():
        path = BASE_DATA_PATH / path
    if not path.exists():
        raise Exception(f"Source {db.source} does not exist")
    logger.info("Importing database ")
    bw_importer = get_bw_importer(db)
    logger.info(f"Importing {db.name} from '{db.source}' using '{bw_importer.__name__}'")
    # return bw_importer
    imported = bw_importer(str(db.source), db.name)
    imported.apply_strategies()
    # print(type(imported))
    if imported.all_linked:
        imported.write_database()



from enbios2.const import BASE_SCHEMA_PATH
from enbios2.generic.enbios2_logging import get_logger, get_module_name
from enbios2.models.experiment_models import ExperimentData
from pydantic import TypeAdapter
import json

logger = get_logger(get_module_name(__file__))


def create_experiment_schema():
    schema_file_path = BASE_SCHEMA_PATH / "experiment.schema.gen.json"
    schema_file_path.write_text(
        json.dumps(TypeAdapter(ExperimentData).json_schema(),indent=2, ensure_ascii=False),
        encoding="utf-8")
    logger.info(f"Created schema file {schema_file_path}")


if __name__ == "__main__":
    create_experiment_schema()



import json
from pathlib import Path

base = Path("###FOLDER WITH ALL THE RESULTS JSONS")
# grab all the result files, and get the top level results
total_res = []
for index in range(len(list(base.glob("*.json")))):
    print(index)
    file = base / f"{index}.json"
    total_res.append(json.load(file.open())[index]["results"])


# store to results.json file
with open("results_blaaaa.json", "w") as f:
    json.dump(total_res, f)


# cols = ["Scenario"] + list#(total_res[0].keys())



import csv
import itertools
import math
import time
from copy import copy
from dataclasses import asdict
from datetime import timedelta
from pathlib import Path
from tempfile import gettempdir
from typing import Optional, Union, Any, cast

import bw2data as bd
import numpy as np
from bw2data.backends import Activity
from pint import Quantity, UndefinedUnitError, DimensionalityError
from pydantic import ValidationError

from enbios2.base.db_models import BWProjectIndex
from enbios2.base.scenario import Scenario
from enbios2.base.stacked_MultiLCA import StackedMultiLCA
from enbios2.base.unit_registry import ureg
from enbios2.bw2.util import bw_unit_fix, get_activity
from enbios2.ecoinvent.ecoinvent_index import get_ecoinvent_dataset_index
from enbios2.generic.enbios2_logging import get_logger
from enbios2.generic.files import PathLike, ReadPath
from enbios2.generic.tree.basic_tree import BasicTreeNode
from enbios2.models.experiment_models import (
    ExperimentActivityId,
    ExtendedExperimentActivityData,
    ExperimentMethodData,
    ExperimentScenarioData,
    ExperimentData,
    EcoInventSimpleIndex,
    MethodsDataTypes,
    ActivitiesDataTypes,
    ScenarioResultNodeData,
    ExperimentMethodPrepData,
    ActivityOutput,
    SimpleScenarioActivityId,
    Activity_Outputs,
    BWCalculationSetup,
    ExperimentActivityData,
    ExperimentConfig,
)

logger = get_logger(__file__)


class Experiment:
    DEFAULT_SCENARIO_ALIAS = "default scenario"

    def __init__(self, raw_data: Union[ExperimentData, dict]):
        if isinstance(raw_data, dict):
            input_data = ExperimentData(**raw_data)
        else:
            input_data = raw_data
        self.raw_data: ExperimentData = input_data
        # alias to activity

        self._validate_bw_config()
        self._activities: dict[
            str, ExtendedExperimentActivityData
        ] = Experiment._validate_activities(
            self._prepare_activities(self.raw_data.activities),
            self.raw_data.bw_default_database,
        )

        self.raw_data.hierarchy = self._prepare_hierarchy()
        self.hierarchy_root: BasicTreeNode[
            ScenarioResultNodeData
        ] = self.validate_hierarchy(self.raw_data.hierarchy)

        self.methods: dict[str, ExperimentMethodPrepData] = Experiment._validate_methods(
            self._prepare_methods()
        )
        self.scenarios: list[Scenario] = self._validate_scenarios()
        self._lca: Optional[StackedMultiLCA] = None
        self._execution_time: float = float("NaN")

    def _validate_bw_config(self) -> None:
        if self.config.use_k_bw_distributions < 1:
            raise ValueError(
                f"config.use_k_bw_distributions must be greater than 0, "
                f"but is {self.config.use_k_bw_distributions}"
            )

        def validate_bw_project_bw_database(
            bw_project: str, bw_default_database: Optional[str] = None
        ):
            if bw_project not in bd.projects:
                raise ValueError(f"Project {bw_project} not found")
            if bw_project in bd.projects:
                bd.projects.set_current(bw_project)

            if bw_default_database:
                if bw_default_database not in bd.databases:
                    raise ValueError(
                        f"Database {bw_default_database} "
                        f"not found. Options are: {list(bd.databases)}"
                    )

        # print("validate_bw_config***************", self.raw_data.bw_project)
        if isinstance(self.raw_data.bw_project, str):
            validate_bw_project_bw_database(
                self.raw_data.bw_project, self.raw_data.bw_default_database
            )

        else:
            simple_index: EcoInventSimpleIndex = self.raw_data.bw_project
            ecoinvent_index = get_ecoinvent_dataset_index(
                version=simple_index.version,
                system_model=simple_index.system_model,
                type_="default",
            )
            if ecoinvent_index:
                ecoinvent_index = ecoinvent_index[0]
            else:
                raise ValueError(f"Ecoinvent index {self.raw_data.bw_project} not found")

            if not ecoinvent_index.bw_project_index:
                raise ValueError(
                    f"Ecoinvent index {ecoinvent_index}, has not BWProject index"
                )
            bw_project_index: BWProjectIndex = ecoinvent_index.bw_project_index
            validate_bw_project_bw_database(
                bw_project_index.project_name, bw_project_index.database_name
            )

    @staticmethod
    def _prepare_activities(
        activities: ActivitiesDataTypes,
    ) -> list[ExperimentActivityData]:
        raw_activities_list: list[ExperimentActivityData] = []
        if isinstance(activities, list):
            raw_activities_list = activities
            for activity in raw_activities_list:
                activity.orig_id = copy(activity.id)

        elif isinstance(activities, dict):
            for activity_alias, activity in activities.items():
                if activity_alias == activity.alias:
                    raise ValueError(
                        f"Activity in activities-dict declared with alias: "
                        f"'{activity_alias}', "
                        f"different than in the activity.id: '{activity.alias}'"
                    )
                activity.orig_id = copy(activity.id)
                activity.id.alias = activity_alias
                raw_activities_list.append(activity)
        return raw_activities_list

    @staticmethod
    def _validate_activities(
        activities: list[ExperimentActivityData],
        bw_default_database: Optional[str] = None,
        output_required: bool = False,
    ) -> dict[str, ExtendedExperimentActivityData]:
        """
        Check if all activities exist in the bw database, and check if the
        given activities are unique
        In case there is only one scenario, all activities are required to have outputs
        """
        # if activities is a list, convert validate and convert to dict
        default_id_data = ExperimentActivityId(database=bw_default_database)
        activities_map: dict[str, ExtendedExperimentActivityData] = {}
        # validate
        for activity in activities:
            ext_activity: ExtendedExperimentActivityData = Experiment._validate_activity(
                activity, default_id_data, output_required
            )
            # check unique aliases
            if ext_activity.alias in activities_map:
                raise Exception(
                    f"Activity-alias '{ext_activity.alias}' passed more then once. "
                    f"Consider using a dictionary for activities or include "
                    f"'alias' in the ids."
                )
            activities_map[ext_activity.alias] = ext_activity

        # all codes should only appear once
        # unique_activities = set()
        # for ext_activity in activities_map.values():
        #     unique_activities.add((ext_activity.id.database, ext_activity.id.code))
        assert len(activities_map) > 0, "There are no activities in the experiment"
        return activities_map

    @staticmethod
    def _bw_activity_search(activity: ExperimentActivityData) -> Activity:
        """
        Search for the activity in the brightway project
        :param activity:
        :return: brightway activity
        """
        id_ = activity.id
        bw_activity: Optional[Activity] = None
        if id_.code:
            if id_.database:
                bw_activity = bd.Database(id_.database).get_node(id_.code)
            else:
                bw_activity = get_activity(id_.code)
        elif id_.name:
            filters = {}
            search_in_dbs = [id_.database] if id_.database else bd.databases
            for db in search_in_dbs:
                if id_.location:
                    filters["location"] = id_.location
                    search_results = bd.Database(db).search(id_.name, filter=filters)
                else:
                    search_results = bd.Database(db).search(id_.name)
                if id_.unit:
                    search_results = list(
                        filter(lambda a: a["unit"] == id_.unit, search_results)
                    )
                if len(search_results) == 1:
                    bw_activity = search_results[0]
                    break
                elif len(search_results) > 1:
                    activities_str = "\n".join(
                        [f'{str(a)} - {a["code"]}' for a in search_results]
                    )
                    raise ValueError(
                        f"There are more than one activity with the same name, "
                        f"try including  "
                        f"the code of the activity you want to use:\n{activities_str}"
                    )
        if not bw_activity:
            raise ValueError(f"No activity found for {activity.id}")
        return bw_activity

    @staticmethod
    def _validate_activity(
        activity: ExperimentActivityData,
        default_id_attr: ExperimentActivityId,
        required_output: bool = False,
    ) -> "ExtendedExperimentActivityData":
        """
        This method checks if the activity exists in the database by several ways.
        :param activity:
        :param default_id_attr:
        :param required_output:
        :return:
        """

        # get the brightway activity
        bw_activity = Experiment._bw_activity_search(activity)

        # create output: ActivityOutput and default_output_value
        if activity.output:
            if isinstance(activity.output, tuple):
                output = ActivityOutput(
                    unit=activity.output[0], magnitude=activity.output[1]
                )
            else:  # if isinstance(activity.output, ActivityOutput):
                output = activity.output

        else:
            output = ActivityOutput(unit=bw_unit_fix(bw_activity["unit"]), magnitude=1.0)
        default_output_value = Experiment._validate_output(
            output, bw_activity, activity.id
        )

        if not activity.output:
            try:
                activity.output = ActivityOutput(unit=bw_activity["unit"])
            except ValidationError as err:
                raise ValueError(
                    f"Activity {activity.id} has invalid output format: {err}"
                )

        activity_dict = asdict(activity)
        activity_dict["output"] = asdict(output)
        result: ExtendedExperimentActivityData = ExtendedExperimentActivityData(
            **activity_dict,
            bw_activity=bw_activity,
            default_output_value=default_output_value,
        )
        result.id.fill_empty_fields(["alias"], **asdict(default_id_attr))

        result.id.fill_empty_fields(
            ["name", "code", "location", "unit", ("alias", "name")],
            **result.bw_activity.as_dict(),
        )
        if required_output:
            assert activity.output is not None, (
                f"Since there is no scenario, activity output is required: "
                f"{activity.orig_id}"
            )
        return result

    @staticmethod
    def _validate_output(
        target_output: ActivityOutput,
        bw_activity: Activity,
        activity_id: ExperimentActivityId,
    ) -> float:
        """
        validate and convert to the bw-activity unit
        :param target_output:
        :param bw_activity:
        :param activity_id:
        :return:
        """
        try:
            target_quantity: Quantity = (
                ureg.parse_expression(
                    bw_unit_fix(target_output.unit), case_sensitive=False
                )
                * target_output.magnitude
            )
            bw_activity_unit = bw_activity["unit"]
            return target_quantity.to(bw_unit_fix(bw_activity_unit)).magnitude
        except UndefinedUnitError as err:
            logger.error(
                f"Cannot parse output unit '{target_output.unit}'- "
                f"of activity {activity_id}. {err}. "
                f"Consider the unit definition to 'enbios2/base/unit_registry.py'"
            )
            raise Exception(f"Unit error, {err}; For activity: {activity_id}")
        except DimensionalityError as err:
            logger.error(
                f"Cannot convert output of activity {activity_id}. -"
                f"From- \n{target_output}\n-To-"
                f"\n{bw_activity['unit']} (brightway unit)"
                f"\n{err}"
            )
            raise Exception(f"Unit error for activity: {activity_id}")

    def _prepare_methods(
        self, methods: Optional[MethodsDataTypes] = None
    ) -> dict[str, ExperimentMethodData]:
        if not methods:
            methods = self.raw_data.methods
        method_dict: dict[str, ExperimentMethodData] = {}
        if isinstance(methods, dict):
            for method_alias, method in methods.items():
                method_dict[method_alias] = ExperimentMethodData(method, method_alias)
        elif isinstance(self.raw_data.methods, list):
            method_list: list[ExperimentMethodData] = self.raw_data.methods
            for method_ in method_list:
                alias = method_.alias if method_.alias else "_".join(method_.id)
                method__ = ExperimentMethodData(method_.id, alias)
                method_dict[method__.alias_] = method__
        return method_dict

    @staticmethod
    def _validate_method(
        method: ExperimentMethodData, alias: str
    ) -> ExperimentMethodPrepData:
        method.id = tuple(method.id)
        bw_method = bd.methods.get(method.id)
        if not bw_method:
            raise Exception(f"Method with id: {method.id} does not exist")
        if method.alias:
            if method.alias != alias:
                raise Exception(
                    f"Method alias: {method.alias} does not match with "
                    f"the given alias: {alias}"
                )
        else:
            method.alias = alias
        return ExperimentMethodPrepData(
            id=method.id, alias=method.alias, bw_method_unit=bw_method["unit"]
        )

    @staticmethod
    def _validate_methods(
        method_dict: dict[str, ExperimentMethodData]
    ) -> dict[str, ExperimentMethodPrepData]:
        # all methods must exist
        return {
            alias: Experiment._validate_method(method, alias)
            for alias, method in method_dict.items()
        }

    def _has_activity(
        self, alias_or_id: Union[str, ExperimentActivityId]
    ) -> Optional[ExtendedExperimentActivityData]:
        if isinstance(alias_or_id, str):
            activity = self._activities.get(alias_or_id, None)
            return activity
        else:  # isinstance(alias_or_id, ExperimentActivityId):
            for activity in self._activities.values():
                if activity.orig_id == alias_or_id:
                    return activity
            return None

    def get_activity(
        self, alias_or_id: Union[str, ExperimentActivityId]
    ) -> ExtendedExperimentActivityData:
        """
        Get an activity by either its alias or its original "id"
        as it is defined in the experiment data
        :param alias_or_id:
        :return: ExtendedExperimentActivityData
        """
        activity = self._has_activity(alias_or_id)
        if not activity:
            raise ValueError(f"Activity with id {alias_or_id} not found")
        return activity

    def _validate_scenarios(self) -> list[Scenario]:
        """
        :return:
        """

        def validate_activity_id(
            activity_id: Union[str, ExperimentActivityId]
        ) -> SimpleScenarioActivityId:
            activity = self.get_activity(activity_id)
            id_ = activity.id
            assert id_.name and id_.code and id_.alias
            return SimpleScenarioActivityId(name=id_.name, alias=id_.alias, code=id_.code)

        def validate_activities(scenario_: ExperimentScenarioData) -> Activity_Outputs:
            activities = scenario_.activities
            result: dict[SimpleScenarioActivityId, float] = {}

            def convert_output(output) -> ActivityOutput:
                if isinstance(output, tuple):
                    return ActivityOutput(
                        unit=bw_unit_fix(output[0]), magnitude=output[1]
                    )
                else:
                    return output  # type: ignore

            if not activities:
                return result

            scenarios_activities = (
                activities if isinstance(activities, list) else activities.items()
            )
            for activity_id, activity_output in scenarios_activities:
                activity = self.get_activity(activity_id)
                simple_id = validate_activity_id(activity_id)
                output_ = convert_output(activity_output)
                scenario_output = Experiment._validate_output(
                    output_, activity.bw_activity, activity.id
                )
                result[simple_id] = scenario_output
            return result

        def validate_scenario(
            _scenario: ExperimentScenarioData, _scenario_alias: str
        ) -> Scenario:
            """
            Validate one scenario
            :param _scenario:
            :param _scenario_alias:
            :return:
            """
            scenario_activities_outputs: Activity_Outputs = validate_activities(_scenario)
            defined_aliases = [
                output_id.alias for output_id in scenario_activities_outputs.keys()
            ]
            # fill up the missing activities with default values
            for activity in self._activities.values():
                activity_alias = activity.alias
                if activity_alias not in defined_aliases:
                    # print(activity)
                    id_ = SimpleScenarioActivityId(
                        name=str(activity.id.name),
                        code=str(activity.id.code),
                        alias=activity.alias,
                    )
                    scenario_activities_outputs[
                        id_
                    ] = activity.default_output_value  # type: ignore

            resolved_methods: dict[str, ExperimentMethodPrepData] = {}
            if _scenario.methods:
                if isinstance(_scenario.methods, list):
                    for index_, method_ in enumerate(_scenario.methods):
                        if isinstance(method_, str):
                            global_method = self.methods.get(method_)
                            assert global_method
                            resolved_methods[global_method.alias] = global_method
                else:
                    method_dict: dict[str, tuple[str, ...]] = cast(
                        dict[str, tuple[str, ...]], _scenario.methods
                    )
                    for method_alias, method_ in method_dict.items():  # type: ignore
                        md = ExperimentMethodData(id=cast(tuple[str, ...], method_))
                        prep_method = self._validate_method(md, method_alias)
                        resolved_methods[prep_method.alias] = prep_method

            return Scenario(
                experiment=self,  # type: ignore
                alias=_scenario_alias,
                activities_outputs=scenario_activities_outputs,
                methods=resolved_methods,
                result_tree=self.hierarchy_root.copy(),
            )

        raw_scenarios = self.raw_data.scenarios
        scenarios: list[Scenario] = []

        # from Union, list or dict
        if isinstance(raw_scenarios, list):
            raw_list_scenarios: list[ExperimentScenarioData] = raw_scenarios
            for index, _scenario in enumerate(raw_list_scenarios):
                _scenario_alias = (
                    _scenario.alias
                    if _scenario.alias
                    else ExperimentScenarioData.alias_factory(index)
                )
                scenarios.append(validate_scenario(_scenario, _scenario_alias))
        elif isinstance(raw_scenarios, dict):
            raw_dict_scenarios: dict[str, ExperimentScenarioData] = raw_scenarios
            for alias, _scenario in raw_dict_scenarios.items():
                if _scenario.alias is not None and _scenario.alias != alias:
                    assert False, (
                        f"Scenario defines alias as dict-key: {alias} but "
                        f"also in the scenario object: {_scenario.alias}"
                    )
                _scenario.alias = alias
                scenarios.append(validate_scenario(_scenario, alias))
        # undefined scenarios. just one default scenario
        elif not raw_scenarios:
            default_scenario = ExperimentScenarioData()
            scenarios.append(
                validate_scenario(default_scenario, Experiment.DEFAULT_SCENARIO_ALIAS)
            )

        for scenario in scenarios:
            scenario.prepare_tree()
        return scenarios

    def _prepare_hierarchy(self) -> Union[dict, list]:
        return (
            self.raw_data.hierarchy
            if self.raw_data.hierarchy
            else list(self._activities.keys())
        )

    def validate_hierarchy(
        self, hierarchy: Union[dict, list]
    ) -> BasicTreeNode[ScenarioResultNodeData]:
        tech_tree: BasicTreeNode[ScenarioResultNodeData] = BasicTreeNode.from_dict(
            hierarchy, compact=True
        )
        for leaf in tech_tree.get_leaves():
            leaf.temp_data = {"activity": self.get_activity(leaf.name)}
        missing = set(self.activities_aliases) - set(
            n.name for n in tech_tree.get_leaves()
        )
        if missing:
            raise ValueError(f"Activities {missing} not found in hierarchy")
        return tech_tree

    def get_scenario(self, scenario_alias: str) -> Scenario:
        """
        Get a scenario by its alias
        :param scenario_alias:
        :return:
        """
        for scenario in self.scenarios:
            if scenario.alias == scenario_alias:
                return scenario
        raise ValueError(f"Scenario '{scenario_alias}' not found")

    def run_scenario(self, scenario_alias: str) -> dict[str, Any]:
        """
        Run a specific scenario
        :param scenario_alias:
        :return: The result_tree converted into a dict
        """
        return self.get_scenario(scenario_alias).run().as_dict(include_data=True)

    def run(self) -> dict[str, BasicTreeNode[ScenarioResultNodeData]]:
        """
        Run all scenarios. Returns an dict with the scenario alias as key
        and the result_tree as value
        :return: dictionary scenario-alias : result_tree
        """
        methods = [m.id for m in self.methods.values()]
        inventories: list[list[dict[Activity, float]]] = []
        for scenario in self.scenarios:
            scenario.reset_execution_time()
            if scenario.methods:
                raise ValueError(
                    f"Scenario cannot have individual methods. '{scenario.alias}'"
                )
            inventory: list[dict[Activity, float]] = []
            for activity_alias, act_out in scenario.activities_outputs.items():
                bw_activity = scenario.experiment.get_activity(
                    activity_alias.alias
                ).bw_activity
                inventory.append({bw_activity: act_out})
            inventories.append(inventory)

        # run experiment
        start_time = time.time()
        calculation_setup = BWCalculationSetup(
            "experiment", list(itertools.chain(*inventories)), methods
        )

        distribution_results = self.config.use_k_bw_distributions > 1

        results: dict[str, BasicTreeNode[ScenarioResultNodeData]] = {}
        for i in range(self.config.use_k_bw_distributions):
            raw_results = StackedMultiLCA(calculation_setup, distribution_results).results
            scenario_results = np.split(raw_results, len(self.scenarios))
            for index, scenario in enumerate(self.scenarios):
                results[scenario.alias] = scenario.set_results(
                    scenario_results[index],
                    distribution_results,
                    i == self.config.use_k_bw_distributions - 1,
                )
            self._execution_time = time.time() - start_time
        return results

    @property
    def execution_time(self) -> str:
        """
        Get the execution time of the experiment (or all its scenarios) in
        a readable format
        :return: execution time in the format HH:MM:SS
        """
        if not math.isnan(self._execution_time):
            return str(timedelta(seconds=int(self._execution_time)))
        else:
            any_scenario_run = False
            scenario_results = ""
            for scenario in self.scenarios:
                if not math.isnan(scenario._execution_time):
                    any_scenario_run = True
                    scenario_results += f"{scenario.alias}: {scenario.execution_time}\n"
            if any_scenario_run:
                return scenario_results
            else:
                return "not run"

    def results_to_csv(
        self,
        file_path: PathLike,
        scenario_alias: Optional[str] = None,
        level_names: Optional[list[str]] = None,
        include_method_units: bool = True,
    ):
        """
        Turn the results into a csv file. If no scenario name is given,
        it will export all scenarios to the same file,
        :param file_path:
        :param scenario_alias: If no scenario name is given, it will
        export all scenarios to the same file,
            with an additional column for the scenario alias
        :param level_names: (list of strings) If given, the results will be
        exported with the given level names
        :param include_method_units:  (Include the units of the methods in the header)
        :return:
        """
        if scenario_alias:
            scenario = self.get_scenario(scenario_alias)
            scenario.results_to_csv(
                file_path,
                level_names=level_names,
                include_method_units=include_method_units,
            )
            return
        else:
            header = []
            all_rows: list = []
            for scenario in self.scenarios:
                temp_file_name = gettempdir() + f"/temp_scenario_{scenario.alias}.csv"
                scenario.results_to_csv(
                    temp_file_name,
                    level_names=level_names,
                    include_method_units=include_method_units,
                )
                rows = ReadPath(temp_file_name).read_data()
                rows[0]["scenario"] = scenario.alias
                if not all_rows:
                    header = list(rows[0].keys())
                    header.remove("scenario")
                    header.insert(0, "scenario")
                all_rows.extend(rows)
                if (temp_file := Path(temp_file_name)).exists():
                    temp_file.unlink()
            with Path(file_path).open("w", newline="") as csvfile:
                writer = csv.DictWriter(csvfile, header)
                writer.writeheader()
                writer.writerows(all_rows)

    def result_to_dict(self, include_output: bool = True) -> list[dict[str, Any]]:
        """
        Get the results of all scenarios as a list of dictionaries as dictionaries
        :param include_output: Include the output of each activity in the tree
        :return:
        """
        return [
            scenario.result_to_dict(include_output=include_output)
            for scenario in self.scenarios
        ]

    @property
    def config(self) -> ExperimentConfig:
        """
        get the config of the experiment
        :return:
        """
        return self.raw_data.config

    def __repr__(self):
        return (
            f"Experiment: (call info() for details)\n"
            f"Activities: {len(self._activities)}\n"
            f"Methods: {len(self.methods)}\n"
            f"Hierarchy (depth): {self.hierarchy_root.depth}\n"
            f"Scenarios: {len(self.scenarios)}\n"
        )

    @property
    def activities_aliases(self) -> list[str]:
        return list(self._activities.keys())

    @property
    def method_aliases(self) -> list[str]:
        return list(self.methods.keys())

    @property
    def scenario_aliases(self) -> list[str]:
        return list([s.alias for s in self.scenarios])

    def info(self) -> str:
        """
        Information about the experiment
        :return:
        """
        activity_rows: list[str] = []
        for activity_alias, activity in self._activities.items():
            activity_rows.append(f"  {activity.alias} - {activity.id.name}")
        activity_rows_str = "\n".join(activity_rows)
        methods_str = "\n".join([f" {m.id}" for m in self.methods.values()])
        return (
            f"Experiment: \n"
            f"Activities: {len(self._activities)}\n"
            f"{activity_rows_str}\n"
            f"Methods: {len(self.methods)}\n"
            f"{methods_str}\n"
            f"Hierarchy (depth): {self.hierarchy_root.depth}\n"
            f"Scenarios: {len(self.scenarios)}\n"
        )



from sentence_transformers import SentenceTransformer


class NLPModels:
    """
    class to store static model
    """
    sentence_model = None

    @classmethod
    def get_sentence_model(clzz, model_name: str = 'all-MiniLM-L6-v2'):
        """
        load model if not loaded
        :param model_name:
        :return:
        """
        if not clzz.sentence_model:
            print("loading model")
            clzz.sentence_model = SentenceTransformer(model_name)
        return clzz.sentence_model



import json
from pathlib import Path

import numpy
import torch
from bw2data.backends import Activity, ActivityDataset, ExchangeDataset
from sentence_transformers import SentenceTransformer
from torch_geometric.nn import GCNConv
from torch_geometric.data import Data

import bw2data
from tqdm import tqdm

from enbios2.experiment.nlp.models import NLPModels

bw2data.projects.set_current("ecoinvent")
db = bw2data.Database("cutoff_3.9.1_default")
# act: Activity = db.random()

feature_path = Path("features")

# here we create node features from the name and the product.
# we store them as jsonl arrays in the features folder

model = NLPModels.get_sentence_model()


name_feature_files = (feature_path / "names.jsonl").open("w")
# product_feature_files = (feature_path / "product.jsonl").open("w")

default_len = len(model.encode("cool"))
for act in tqdm(ActivityDataset.select().order_by(ActivityDataset.id)):
    name_feature_files.write(json.dumps(model.encode(act.name).tolist()) + "\n")
    # if act.product:
    #     product_feature_files.write(json.dumps(model.encode(act.product).tolist()) + "\n")
    # else:
    #     product_feature_files.writelines(json.dumps(numpy.zeros(default_len).tolist()) + "\n")
name_feature_files.close()
# product_feature_files.close()

# Now lets create edges and their features
# ids are indices... note that our ids start with 0. while those of bw (sqllite database start with 1)
# we store them in a n x k array, where k is min 2, first 2. are in/out nodes
code2idMap = [act.code for act in tqdm(ActivityDataset.select().order_by(ActivityDataset.id))]
edges_file = (feature_path / "edges.jsonl").open("w")
for e in tqdm(ExchangeDataset.select().order_by(ExchangeDataset.id)):
    edges_file.write(json.dumps([
        code2idMap.index(e.output_code),
        code2idMap.index(e.input_code),
        e.data.get("amount",0)
    ]) + "\n")
edges_file.close()

# X = torch.tensor(node_features, dtype=torch.float)
# edges = torch.tensor(edge_list, dtype=torch.long)
# edge_values = torch.tensor(edge_values, dtype=torch.float)
# scores = torch.tensor(node_scores, dtype=torch.float)
#
# # Create a PyG Data object
# data = Data(x=X, edge_index=edges.t().contiguous(), edge_attr=edge_values, y=scores)


import numpy as np
import torch

# Assume we have some node data in the variables node_features
# node_features is a numpy array of shape (num_nodes, num_features)



from bw2data.backends import Activity


def get_connections(activity: Activity):
    return {
        "biosphere": list(activity.biosphere()),
        "technosphere": list(activity.technosphere()),
        "consumers": list(activity.consumers()),
    }


def connections_size(conns: dict[str,any]):
    return {k: len(v) for k,v in conns.items()}


import json
import logging.config
import os
from logging import getLogger
from pathlib import Path
from typing import Optional, Any

from enbios2.const import BASE_DATA_PATH, PROJECT_PATH


default_log_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "raw": {"format": "%(message)s"},
        "simple": {"format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"},
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "simple",
            "stream": "ext://sys.stdout",
        }
    },
    "loggers": {},
    "root": {"level": "INFO", "handlers": ["console"]},
}

new_logger_default_config = {
    "level": "DEBUG",
    "handlers": ["console"],
    "propagate": False,
}


class EnbiosLogger:
    initialized = False
    log_config_file: Optional[Path] = None
    config_data: dict[str, Any] = {}

    @classmethod
    def init_logger(cls):
        cls.log_config_file = Path(BASE_DATA_PATH) / "logging.json"
        # check if  exists
        if not cls.log_config_file.exists():
            print(f"Creating logging config file at: {cls.log_config_file}")
            # if not, create it
            cls.log_config_file.parent.mkdir(parents=True, exist_ok=True)
            cls.log_config_file.write_text(
                json.dumps(default_log_config, ensure_ascii=False, indent=2),
                encoding="utf-8",
            )

        cls.reload_config()
        # shutoff bw2calc logger
        getLogger("bw2calc").setLevel(logging.WARNING)

    @classmethod
    def reload_config(cls):
        cls.config_data = json.loads(cls.log_config_file.read_text(encoding="utf-8"))
        logging.config.dictConfig(cls.config_data)

    @classmethod
    def add_logger(cls, name):
        cls.config_data["loggers"][name] = new_logger_default_config
        # write to file
        cls.log_config_file.write_text(
            json.dumps(cls.config_data, ensure_ascii=False, indent=2), encoding="utf-8"
        )

    @classmethod
    def get_or_create_logger(cls, name: str) -> logging.Logger:
        if not cls.initialized:
            cls.init_logger()
        if name not in cls.config_data["loggers"]:
            cls.add_logger(name)
            logging.config.dictConfig(cls.config_data)
        return logging.getLogger(name)


def get_module_name(file_path: str) -> str:
    """Get the module name based on the file's location in the project."""
    relative_path = os.path.relpath(file_path, PROJECT_PATH)
    module_name = os.path.splitext(relative_path)[0]
    module_name = module_name.replace(os.sep, ".")
    return module_name


def get_logger(file_path: str) -> logging.Logger:
    """
    Get a logger for the given module, based on its file path.
    use like this:
    Takes the logging config from logging.json.
    Creates a new entry in that file if it does not exist yet with
    'new_logger_default_config'.
    logger = get_logger(__file__)
    :param file_path:  absolute file path
    :return: logger for the module...
    """
    return EnbiosLogger.get_or_create_logger(get_module_name(file_path))




import bw2data as bd
from bw2calc import LCA
from bw2calc.monte_carlo import MonteCarloLCA, MultiMonteCarlo

bd.projects.set_current("ecoinvent_391")  # Select your project
ei = bd.Database("ecoinvent_391_cutoff")

act = ei.get_node('a8fe0b37705fe611fac8004ca6cb1afd')
act2 = ei.get_node('413bc4617794c6e8b07038dbeca64adb')

method = ('CML v4.8 2016', 'climate change', 'global warming potential (GWP100)')

# Fer un LCA simple

for i in range(1):
    lca = LCA({act: 1, act2: 3}, method, use_distributions=True)
    lca.lci()
    lca.lcia()
    print(lca.score)

demands1 = [{act.key: 1}]
print(act.key)
demands = [{act: 1,
            act2: 10}]


# mc = MonteCarloLCA(demands[0], data_objs=[])
# import pickle
# pickle.dumps(mc)

mc = MultiMonteCarlo(demands1, method=method)
#
mc.calculate()



"""
This requires postgresql and the pgvector extension to be installed.
https://github.com/pgvector/pgvector
sqlalchemy, sqlmodel, psycopg2, and sentence_transformers
"""
from typing import List, Optional

from pgvector.sqlalchemy import Vector
from sqlalchemy import Column, Integer
from sqlalchemy.dialects.postgresql import ARRAY
from sqlmodel import SQLModel, Field, create_engine


class Document(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    content: str
    ext_ids:list[int] = Field([], sa_column=Column(ARRAY(Integer)))
    embedding: List[float] = Field(sa_column=Column(Vector(384)))
    norm_embedding: Optional[List[float]] = Field(sa_column=Column(Vector(384)))
    cluster : Optional[int]
    embedding_2d: Optional[List[float]] = Field(sa_column=Column(Vector(2)))
    norm_embedding_2d: Optional[List[float]] = Field(sa_column=Column(Vector(2)))


def init_engine(user: str, password: str, host: str, db_name: str):
    return create_engine(f'postgresql://{user}:{password}@{host}/{db_name}')


def create_database(db_name, user, password, host='localhost'):
    # connect to the default 'postgres' database
    engine = create_engine(f'postgresql://{user}:{password}@{host}/postgres')

    # get a raw DBAPI connection
    conn = engine.raw_connection()
    try:
        conn.set_isolation_level(0)  # AUTOCOMMIT
        conn.cursor().execute(f"CREATE DATABASE {db_name};")
    finally:
        conn.set_isolation_level(1)  # reset isolation level
        conn.close()

    # now connect to the new database and create extension
    # todo, following line required?
    engine = create_engine(f'postgresql://{user}:{password}@{host}/{db_name}')
    conn = engine.raw_connection()
    try:
        conn.set_isolation_level(0)  # AUTOCOMMIT
        conn.cursor().execute("CREATE EXTENSION IF NOT EXISTS vector;")
    finally:
        conn.set_isolation_level(1)  # reset isolation level
        conn.close()

def create_vector_db():
    create_database("activities_vectors", user_name, password)

def reset_vector_db():
    SQLModel.metadata.drop_all(engine)
    SQLModel.metadata.create_all(engine)


user_name = "postgres"
password = "bbe90ce51d87635c7354445e6887c73a"

engine = init_engine(user_name, password, "localhost", "activities_vectors")
SQLModel.metadata.create_all(engine)


"""
convert spold files to json files and create a schema for them and a set of units.
Do a deepdiff between the schemas
"""

import json
from os import listdir
from pathlib import Path
from typing import Optional

import xmltodict
from deepdiff.diff import DeepDiff
from genson import SchemaBuilder
from tqdm import tqdm

from enbios2.const import BASE_DATA_PATH

ecoinvent_folder = BASE_DATA_PATH / "ecoinvent"
from jsonpath_ng import parse


def get_output_folder_for_spold(spold_folder: Path) -> Path:
    output_folder = spold_folder.relative_to(ecoinvent_folder)
    # should be ecoinventJson/ecoinvent 3.9.1_cutoff_ecoSpold02/datasets
    output_folder = BASE_DATA_PATH / "ecoinventJson" / output_folder
    return output_folder


def convert_folder(spold_folder: Path) -> Path:
    output_folder = get_output_folder_for_spold(spold_folder)
    output_folder.mkdir(parents=True, exist_ok=True)
    for file in tqdm(spold_folder.glob("*.spold")):
        # print(file)
        rel = file.relative_to(spold_folder)
        # print(rel)
        with open(file, "r") as f:
            data: dict = xmltodict.parse(f.read())
            # write to output_folder
            with open(output_folder / rel.with_suffix(".json"), "w", encoding="utf-8") as f2:
                f2.write(json.dumps(data))
    return output_folder


def build_schema_from_spold(data: dict, prev_builder: SchemaBuilder = None) -> SchemaBuilder:
    if prev_builder:
        builder = prev_builder
    else:
        builder = SchemaBuilder()
        builder.add_schema({"type": "object", "properties": {}})

    builder.add_object(data)
    return builder


def read_units(data: dict, prev_matches: set[str] = None) -> set[str]:
    jsonpath_expr = parse(
        '$.ecoSpold.["activityDataset", "childActivityDataset"][*].flowData.intermediateExchange[*].unitName[*]."#text"')
    res = set(match.value for match in jsonpath_expr.find(data))
    if prev_matches:
        prev_matches.update(res)
        return prev_matches
    else:
        return res


def data_iterator_funcs(output_folder: Path):
    print(output_folder)
    files = listdir(output_folder)

    unit_set = set()
    builder: Optional[SchemaBuilder] = None
    for file in tqdm(files):
        with open(output_folder / file, "r", encoding="utf-8") as f:
            data = json.loads(f.read())
            unit_set = read_units(data, unit_set)
            builder = build_schema_from_spold(data, builder)

    (output_folder.parent / "schema.json").write_text(json.dumps(builder.to_schema(), indent=2),
                                                      encoding="utf-8")
    (output_folder.parent / "units.json").write_text(json.dumps(list(unit_set), indent=2),
                                                     encoding="utf-8")


sources = ["ecoinvent 3.9.1_cutoff_ecoSpold02", "ecoinvent 3.9.1_cutoff_lcia_ecoSpold02"]

schema_map = {}
for source in sources:
    # spold_folder = ecoinvent_folder / "ecoinvent 3.9.1_cutoff_ecoSpold02/datasets"
    spold_folder = ecoinvent_folder / f"{source}/datasets"
    output_folder = get_output_folder_for_spold(spold_folder)
    if not output_folder.exists():
        output_folder = convert_folder(spold_folder)

    data_iterator_funcs(output_folder)

    # compare the 2 schemas
    diff = DeepDiff(*schema_map.values(), ignore_order=True)




from enbios2.const import BASE_SCHEMA_PATH
from enbios2.generic.enbios2_logging import get_logger

logger = get_logger(__file__)

def run_all():
    BASE_SCHEMA_PATH.mkdir(parents=True, exist_ok=True)
    from enbios2.dev.create_experiment_schema import create_experiment_schema

    create_experiment_schema()
    logger.warning("Tech tree schema not generated")
    # create_tech_tree_node_schema()


if __name__ == "__main__":
    run_all()


from csv import DictReader
from os import PathLike
from typing import TYPE_CHECKING, Union

from flatten_dict import unflatten
from frictionless import Schema, Resource, validate, system
from frictionless.fields import NumberField, StringField

if TYPE_CHECKING:
    from enbios2.base.experiment import Experiment


from enbios2.generic.files import ReadPath
from enbios2.generic.tree.basic_tree import BasicTreeNode
from enbios2.models.experiment_models import (
    ExperimentDataIO,
    ExperimentData,
    ActivitiesDataRows,
    ExperimentMethodData,
)

activities_schema = Schema(
    fields=[
        StringField(name="alias"),
        StringField(name="database"),
        StringField(name="code"),
        StringField(name="name"),
        StringField(name="location"),
        StringField(name="unit"),
        NumberField(name="output.value", float_number=True),
        StringField(name="output.unit"),
    ]
)

activities_unflatten_map = {
    "alias": "id.alias",
    "database": "id.database",
    "code": "id.code",
    "name": "id.name",
    "location": "id.location",
    "unit": "id.unit",
    "output.value": "output.value",
    "output.unit": "output.unit",
}

methods_schema = Schema(
    fields=[
        StringField(name="alias"),
        StringField(name="id"),
        StringField(name="id0"),
        StringField(name="id1"),
        StringField(name="id2"),
        StringField(name="id3"),
    ]
)


# methods_fields = ["alias", "id", "id0", "id1", "id2", "id3"]


def unflatten_data(data: dict, structure_map: dict):
    res = {}
    for key, value in data.items():
        if key in structure_map:
            res[structure_map[key]] = value
        else:
            res[key] = value
    print(res)
    return unflatten(res, splitter="dot")


def parse_method_row(row: dict) -> ExperimentMethodData:
    id_: list[str] = []
    if row_id := row.get("id"):
        id_ = row_id.split(",")
    else:
        for i in range(4):
            if row_id := row.get(f"id{i}"):
                id_.append(row_id)
            else:
                break
    return ExperimentMethodData(alias=row.get("alias"), id=tuple(id_))


def read_experiment_io(data: Union[dict, ExperimentDataIO]) -> "Experiment":
    from base.experiment import Experiment

    if isinstance(data, dict):
        exp_io = ExperimentDataIO(**data)
    else:
        exp_io = data

    raw_experiment_data = {"bw_project": exp_io.bw_project}

    def get_abs_path(path: Union[str, PathLike]) -> ReadPath:
        if exp_io.config.base_directory:
            return ReadPath(exp_io.config.base_directory) / path
        else:
            return ReadPath(path)

    if isinstance(exp_io.activities, str) or isinstance(exp_io.activities, PathLike):
        activities_file: ReadPath = get_abs_path(exp_io.activities)
        if activities_file.suffix == ".json":
            data = activities_file.read_data()
            raw_experiment_data["activities"] = data
        elif activities_file.suffix == ".csv":
            with system.use_context(trusted=True):
                resource: Resource = Resource(
                    path=activities_file.as_posix(), schema=activities_schema
                )
                report = validate(resource)
                if not report.valid:
                    raise Exception(
                        f"Invalid activities file: {exp_io.activities}; "
                        f"errors: {report.task.errors}"
                    )
                raw_experiment_data["activities"] = ActivitiesDataRows(
                    list(
                        (
                            unflatten_data(r, activities_unflatten_map)
                            for r in resource.read_rows()
                        )
                    )
                )
    else:
        raw_experiment_data["activities"] = exp_io.activities

    if isinstance(exp_io.methods, str) or isinstance(exp_io.methods, PathLike):
        methods_file: ReadPath = get_abs_path(exp_io.methods)
        if methods_file.suffix == ".json":
            data = methods_file.read_data()
            raw_experiment_data["methods"] = data
        if methods_file.suffix == ".csv":
            with system.use_context(trusted=True):
                resource: Resource = Resource(
                    path=methods_file.as_posix(), schema=methods_schema
                )
                report = validate(resource)
                if not report.valid:
                    raise Exception(
                        f"Invalid methods file: {exp_io.methods}; "
                        f"errors: {report.task.errors}"
                    )
                raw_experiment_data["methods"] = list(
                    (parse_method_row(r) for r in resource.read_rows())
                )
    else:
        raw_experiment_data["methods"] = exp_io.methods

    if isinstance(exp_io.hierarchy, str) or isinstance(exp_io.hierarchy, PathLike):
        hierarchy_file: ReadPath = get_abs_path(exp_io.hierarchy)
        if hierarchy_file.suffix == ".json":
            data = hierarchy_file.read_data()
            raw_experiment_data["hierarchy"] = data
        else:
            raise Exception(f"Invalid hierarchy file: {exp_io.hierarchy}")

    if isinstance(exp_io.scenarios, str) or isinstance(exp_io.scenarios, PathLike):
        scenario_file: ReadPath = get_abs_path(exp_io.scenarios)
        if scenario_file.suffix == ".json":
            data = scenario_file.read_data()
            raw_experiment_data["scenario"] = data
        else:
            raise Exception(f"Invalid scenario file: {exp_io.scenarios}")

    return Experiment(raw_experiment_data)


if __name__ == "__main__":
    scenario_data = {
        "bw_project": "uab_bw_ei39",
        "activities_config": {"default_database": "ei391"},
        "config": {
            "base_directory": "/mnt/SSD/projects/LIVENLab/enbios2/"
            "data/test_data/experiment_separated/a/"
        },
        "activities": "single_activity.csv",
        "methods": "methods.csv",
        "hierarchy": {"energy": ["single_activity"]},
    }
    exp: ExperimentData = read_experiment_io(scenario_data)

    reader = DictReader(
        open(
            "/mnt/SSD/projects/LIVENLab/enbios2/data/templates/experiment_activity_data.csv"
        )
    )

    node = BasicTreeNode.from_compact_dict(scenario_data["hierarchy"])



from typing import Optional, Union

from numpy import ndarray
from pandas import DataFrame
from sklearn.preprocessing import MinMaxScaler

from enbios2.base.experiment import Experiment


class ResultsSelector:
    def __init__(
        self,
        experiment: Experiment,
        scenarios: Optional[list[str]] = None,
        methods: Optional[list[str]] = None,
    ):
        self.experiment = experiment

        all_aliases = [sc.alias for sc in self.experiment.scenarios]
        if scenarios is not None:
            for scenario in scenarios:
                if scenario not in all_aliases:
                    raise ValueError(f"Scenario {scenario} not found in experiment")
            self.scenarios = scenarios
        else:
            self.scenarios = all_aliases

        if methods is not None:
            for method in methods:
                if method not in self.experiment.methods.keys():
                    raise ValueError(f"Method {method} not found in experiment")
            self.methods = methods
        else:
            self.methods: list[str] = [
                m for m in self.experiment.methods.keys()
            ]  # type: ignore

        self._complete_df = None
        self._base_df = None

    @staticmethod
    def get_result_selector(
        result_selector: Union[Experiment, "ResultsSelector"],
        scenarios: Optional[list[str]] = None,
        methods: Optional[list[str]] = None,
    ) -> "ResultsSelector":
        if isinstance(result_selector, Experiment):
            return ResultsSelector(result_selector, scenarios=scenarios, methods=methods)
        return result_selector

    def check_special_df(self, df: DataFrame):
        if not self.base_df.shape == df.shape:
            raise ValueError(
                f"Dataframe shape {df.shape} does not match base dataframe shape "
                f"{self.base_df.shape}"
            )
        if not list(self.base_df.dtypes) == list(df.dtypes):
            raise ValueError(
                f"Dataframe dtypes {df.dtypes} do not match base dataframe dtypes "
                f"{self.base_df.dtypes}"
            )

    @property
    def complete_df(self) -> DataFrame:
        if self._complete_df is None:
            data = [
                {"scenario": scenario}
                | {
                    k: v
                    for k, v in self.experiment.get_scenario(
                        scenario
                    ).result_tree.data.results.items()
                }
                for scenario in self.experiment.scenario_aliases
            ]

            self._complete_df = DataFrame(data)
        return self._complete_df

    @property
    def base_df(self) -> DataFrame:
        if self._base_df is None:
            data = [
                {"scenario": scenario}
                | {
                    k: v
                    for k, v in self.experiment.get_scenario(
                        scenario
                    ).result_tree.data.results.items()
                    if k in self.methods
                }
                for scenario in self.scenarios
            ]

            self._base_df = DataFrame(data)
        return self._base_df

    def normalized_df(self, normalize_with_all_scenarios: bool = True) -> DataFrame:
        used_df = self.complete_df if normalize_with_all_scenarios else self.base_df
        columns = used_df.columns[1:]
        normalized_df = DataFrame()
        normalized_df["scenario"] = used_df["scenario"]
        for column in columns:
            normalized_df[column] = MinMaxScaler().fit_transform(
                used_df[column].to_numpy().reshape(-1, 1)
            )

        if normalize_with_all_scenarios:
            # delete all rows where scenario is not in self.scenarios
            normalized_df = normalized_df[normalized_df["scenario"].isin(self.scenarios)]
        return normalized_df

    def method_label_names(
        self, short: bool = True, include_unit: bool = True
    ) -> list[str]:
        exp_methods = self.experiment.methods
        return [
            (
                (exp_methods[method].id[-1] if short else "\n".join(exp_methods))
                + ("\n" + exp_methods[method].bw_method_unit if include_unit else "")
            )
            for method in self.methods
        ]

    def compare_to_baseline(self, baseline_data: ndarray):
        assert len(baseline_data) == len(self.methods), (
            f"Baseline data must have the same length as the number of methods "
            f"({len(self.methods)})"
        )
        # Create a copy of the original dataframe, without modifying it in place
        baseline_df = self.base_df.copy()
        # Use loc to select the columns related to methods
        # and divide them by baseline_data
        baseline_df[self.methods] = self.base_df[self.methods].div(baseline_data, axis=1)
        # Set the dtypes back to their original types
        original_dtypes = self.base_df[self.methods].dtypes
        for col, dtype in original_dtypes.items():
            # print(col, dtype)
            baseline_df[col] = baseline_df[col].astype(dtype)
        return baseline_df.reset_index(drop=True)

    def collect_tech_results(self, node_aliases: list[str]):
        df = DataFrame()
        for scenario in self.scenarios:
            scenario_results = self.experiment.get_scenario(scenario).result_tree
            for node_alias in node_aliases:
                node = scenario_results.find_subnode_by_name(node_alias)
                assert node is not None
                # add a row
                df = df._append(
                    {
                        "scenario": scenario,
                        "tech": node_alias,
                    }
                    | {
                        method: value
                        for method, value in node.data.results.items()
                        if method in self.methods
                    },
                    ignore_index=True,
                )

        return df



from pathlib import Path
from typing import Optional

from peewee import (
    Model,
    TextField,
    FloatField,
    BooleanField,
    SqliteDatabase,
    ForeignKeyField,
)
from playhouse.shortcuts import model_to_dict
from playhouse.sqlite_ext import FTSModel, JSONField

from enbios2.base.db_fields import TupleJSONField, PathField
from enbios2.const import MAIN_DATABASE_PATH


class MainDatabase(Model):
    pass

    class Meta:
        database = SqliteDatabase(MAIN_DATABASE_PATH)


class EcoinventDataset(MainDatabase):
    """
    Ecoinvent datasets model
    """

    version = TextField()  # should have validation // see ecoinvent_consts
    system_model = TextField()  # cutoff, consequential, apos
    type = TextField(default="default")  # ecoinvent_dataset_types
    xlsx = BooleanField(default=False)
    directory: Path = PathField(null=True)  # Path typehint, so that static checker chills
    identity = TextField(unique=True)

    _V391 = "3.9.1"
    _SM_CUTOFF = "cutoff"
    _SM_CONSEQUENTIAL = "consequential"
    _SM_APOS = "apos"
    _valid_ecoinvent_versions = {"3.8", "3.9", "3.9.1"}
    _valid_ecoinvent_system_models = {"cutoff", "consequential", "apos"}
    _valid_ecoinvent_datatypes = {"default", "lci", "lcia"}

    @property
    def bw_project_index(self) -> Optional["BWProjectIndex"]:
        return (
            BWProjectIndex.select()
            .where(BWProjectIndex.ecoinvent_dataset == self)
            .get_or_none()
        )

    class Meta:
        table_name = "ecoinvent_dataset"

    def __init__(self, *args, **kwargs):
        super(EcoinventDataset, self).__init__(*args, **kwargs)
        self.identity = (
            f"{self.system_model}_{self.version}_{self.type}"
            f"{'_xlsx' if self.xlsx else ''}"
        )

    def validate(self):
        check_fields = [
            ("version", EcoinventDataset._valid_ecoinvent_versions),
            ("system_model", EcoinventDataset._valid_ecoinvent_system_models),
            ("type", EcoinventDataset._valid_ecoinvent_datatypes),
        ]
        for field, valid_values in check_fields:
            if getattr(self, field) not in valid_values:
                raise ValueError(
                    f"EcoinventIndex entry '{field}' is not valid: "
                    f"'{getattr(self, field)}'. Must be of {valid_values}"
                    f" ('{self.identity}'/ {self.directory})"
                )

    @classmethod
    def exising_or_new(cls, *args, **kwargs) -> "EcoinventDataset":
        index = EcoinventDataset(*args, **kwargs)
        if EcoinventDataset.exists(index.identity):
            return EcoinventDataset.get_or_none(
                EcoinventDataset.identity == index.identity
            )
        return index

    def save(self, *args, **kwargs):
        self.validate()
        super(EcoinventDataset, self).save(*args, **kwargs)

    @classmethod
    def exists(cls, identity: str) -> bool:
        """
        Check if the given identity exists in the database
        :param identity: identity string
        :return: True, if exists
        """
        return cls.select().where(cls.identity == identity).exists()

    @property
    def dataset_path(self) -> Path:
        """
        get the definite path of the dataset (spold files) or Excel file
        :return: Path of 'datasets' within the folder or Excel file
        """
        if self.xlsx:
            return next(self.directory.glob("*.xlsx"))
        else:
            return self.directory / "datasets"

    def __repr__(self):
        index_str = str(self.bw_project_index)
        return (
            f"EcoinventDataset: {self.identity} "
            f"({index_str if self.bw_project_index else 'no BW2 project index'})"
        )

    def __str__(self):
        return self.__repr__()

    @staticmethod
    def dump_database(
        entries: Optional[list["EcoinventDataset"]] = None,
        redact_dir: Optional[bool] = True,
    ) -> list[dict]:
        """
        Dump the database to a JSON file
        :return:
        """
        res = []
        for entry in entries if entries else EcoinventDataset.select():
            d = model_to_dict(entry)
            if redact_dir:
                d["directory"] = "***/" + entry.directory.name
            if bwp := entry.bw_project_db.get_or_none():
                d["bw_project"] = [bwp.project_name, bwp.database_name]
            res.append(d)
        return res


class EcoinventResolvedDataset(MainDatabase):
    """
    Metadata table for Ecoinvent databases
    """

    name = TextField()
    path = TextField()
    db_type = TextField()
    metadata = JSONField()
    ecoinvent_dataset = ForeignKeyField(
        EcoinventDataset, backref="resolved_dataset", unique=True
    )

    class Meta:
        table_name = "ecoinvent_resolved_dataset"


class BWProjectIndex(MainDatabase):
    project_name = TextField()
    database_name = TextField()
    ecoinvent_dataset = ForeignKeyField(
        EcoinventDataset, backref="bw_project_db", unique=True
    )

    class Meta:
        table_name = "bw_project_index"

    @property
    def get_ecoinvent_dataset(self) -> Optional[EcoinventDataset]:
        return EcoinventDataset.get_or_none(EcoinventDataset == self.ecoinvent_dataset)

    def delete(self):
        self.delete()

    def __repr__(self):
        return (
            f"BWProjectIndex: {self.project_name} - "
            f"{self.database_name} ({self.ecoinvent_dataset})"
        )

    def __str__(self):
        return f"BWProjectIndex: {self.project_name} - {self.database_name}"


class EcoinventDatabaseActivity(Model):
    """
    Main table for the 2 resolved Ecoinvent databases (LCI, LCIA)
    """

    code = TextField(unique=True)
    name = TextField(index=True)
    location = TextField()
    product = TextField()
    product_unit = TextField()
    amount = FloatField()
    data = TupleJSONField()

    class Meta:
        pass


class ExchangeInfo(Model):
    """
    Ecoinvent LCI Index
    """

    exchange = TextField()
    compartment = TextField()
    sub_compartment = TextField()
    unit = TextField()

    class Meta:
        pass


class ImpactInfo(Model):
    """
    Ecoinvent LCIA Index
    """

    method = TextField()
    category = TextField()
    indicator = TextField()
    unit = TextField()

    class Meta:
        pass


class BW_Activity(Model):
    """
    experimental FTS database
    """

    code = TextField()
    database = TextField()
    name = TextField()
    location = TextField(null=True)
    location_name = TextField(null=True)
    product = TextField(null=True)
    type = TextField()
    # from data
    # categories = TextField()
    comment = TextField(null=True)
    classification = TextField(null=True)
    # synonyms = TextField()
    unit = TextField(null=True)
    reference_product = TextField(null=True)


class FTS_BW_ActivitySimple(FTSModel):
    """
    experiment full text search database
    """

    name = TextField()
    product = TextField()
    synonyms = TextField()
    location_name = TextField()
    comment = TextField()
    content = TextField()





from typing import Optional

from genson import SchemaBuilder


def iterative_model_build(data, builder: Optional[SchemaBuilder] = None) -> SchemaBuilder:
    """
    Build a schema from a json object. Uses existing builder if provided. Otherwise creates a new one.
    :param data:
    :param builder:
    :return:
    """
    if builder is None:
        builder = SchemaBuilder()
        builder.add_schema({"type": "object", "properties": {}})
    builder.add_object(data)
    return builder




import logging
import numpy as np

from enbios2.models.experiment_models import BWCalculationSetup

try:
    from bw2data import calculation_setups, get_activity
    from bw2data.backends.proxies import Activity
except ImportError:
    calculation_setups = None

    class Activity:
        pass


from bw2calc.lca import LCA
from bw2calc.utils import wrap_functional_unit

logger = logging.getLogger("bw2calc")


class InventoryMatrices:
    def __init__(self, biosphere_matrix, supply_arrays):
        self.biosphere_matrix = biosphere_matrix
        self.supply_arrays = supply_arrays

    def __getitem__(self, fu_index):
        if fu_index is Ellipsis:
            raise ValueError("Must specify integer indices")

        return self.biosphere_matrix * self.supply_arrays[fu_index]


class StackedMultiLCA:
    """Wrapper class for performing LCA calculations with
    many functional units and LCIA methods.

    Needs to be passed a ``calculation_setup`` name.

    This class does not subclass the `LCA` class, and performs all
    calculations upon instantiation.

    Initialization creates `self.results`, which is a NumPy array of LCA scores,
    with rows of functional units and columns of LCIA methods.
    Ordering is the same as in the `calculation_setup`.

    """

    def __init__(
        self,
        calc_setup: BWCalculationSetup,
        use_distributions: bool = False,
        log_config=None,
    ):
        self.func_units = calc_setup.inv
        self.methods = calc_setup.ia

        self.lca = LCA(
            demand=self.all,
            method=self.methods[0],
            log_config=log_config,
            use_distributions=use_distributions,
        )
        logger.info(
            {
                "message": "Started MultiLCA calculation",
                "methods": list(self.methods),
                "functional units": [wrap_functional_unit(o) for o in self.func_units],
            }
        )
        self.lca.lci()
        self.method_matrices = []
        self.supply_arrays = []
        self.results = np.zeros((len(self.func_units), len(self.methods)))
        for method in self.methods:
            self.lca.switch_method(method)
            self.method_matrices.append(self.lca.characterization_matrix)

        for row, func_unit in enumerate(self.func_units):
            fu_spec, fu_demand = list(func_unit.items())[0]
            if isinstance(fu_spec, int):
                fu = {fu_spec: fu_demand}
            elif isinstance(fu_spec, Activity):
                fu = {fu[0].id: fu[1] for fu in list(func_unit.items())}
            elif isinstance(fu_spec, tuple):
                a = get_activity(fu_spec)
                fu = {a.id: fu[1] for fu in list(func_unit.items())}
            else:
                raise ValueError("Unknown functional unit type")
            self.lca.lci(fu)
            self.supply_arrays.append(self.lca.supply_array)

            for col, cf_matrix in enumerate(self.method_matrices):
                self.lca.characterization_matrix = cf_matrix
                self.lca.lcia_calculation()
                self.results[row, col] = self.lca.score

        self.inventory = InventoryMatrices(self.lca.biosphere_matrix, self.supply_arrays)

    @property
    def all(self):
        """Get all possible databases by merging all functional units"""
        return {key: 1 for func_unit in self.func_units for key in func_unit}



from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Union, Literal

from voluptuous import default_factory

from enum import Enum


class EcoInventVersion(Enum):
    _V391Plus = "3.9.1+"
    _V391 = "3.9.1"
    _V39 = "3.9"
    _V38 = "3.8"
    VERSION_UNDEFINED = "version_undefined"
    NOT_ECOINVENT = "not_ecoinvent"


@dataclass
class BWProject:
    project_name: str
    databases: list["BWProjectDatabase"] = default_factory(list)


@dataclass
class BWProjectDatabase:
    name: str
    source: Union[str, Path]
    importer: Optional[str] = None
    format: Optional[
        Literal[
            "CSV",
            "Ecospold1",
            "Ecoinvent",
            "Excel",
            "Exiobase 3.3.17 hybrid mrio_common_metadata tidy datapackage",
            "Exiobase 3",
            "SimaPro",
            "Ecospold2",
        ]
    ] = None
    ecoinvent_version: Optional[EcoInventVersion] = EcoInventVersion.VERSION_UNDEFINED


def __post_init__(self):
    if self.importer is None and self.format is None:
        raise Exception("Either importer or format must be specified")
    if self.ecoinvent_version == EcoInventVersion._V38:
        import bw2io

        if bw2io.__version__ != (0, 9, "DEV11"):
            raise Exception(
                f"bw2io version must be 0.9.DEV11 for ecoinvent {self.ecoinvent_version}"
            )



import bw2data
from bw2calc import MultiLCA
from bw2data import calculation_setups
from bw2data.backends import Activity

bw2data.projects.set_current("ecoinvent")
db = bw2data.Database("cutoff_3.9.1_default")

if __name__ == "__main__":
    for i in range(10):
        act: Activity = db.random()
        method = list(bw2data.methods)[0]
        single_result = act.lca(method).score
        print(f"{act} - {method}")

        calculation_setups["multiLCATest"] = {
            "inv": [{act: 1}],
            "ia": [method]
        }

        calculation_setups.flush()
        multi_result = MultiLCA("multiLCATest").results[0][0]

        print(single_result)
        print(multi_result)
        print(abs(single_result - multi_result))


import pickle
from pathlib import Path

import bw2data
import networkx as nx
from bw2data.backends import ActivityDataset, ExchangeDataset
from networkx import dominating_set
from tqdm import tqdm

from enbios2.bw2.project_index import set_bw_current_project

# Create an empty graph
G = nx.DiGraph()

set_bw_current_project("cutoff", "3.9.1")
# print(list(bw2data.databases))
# random_act = db.random()
db = bw2data.Database("cutoff_3.9.1_default")
activities = ActivityDataset.select()

# g_node_data = []
for act in tqdm(activities):
    # data = model_to_dict(act)
    # data["location_name"] = geo_code2_name.get(act.location, "")
    # G.add_nodes_from([(act.code, data)]
    # g_node_data.append((act.code, {"name": act.name}))
    G.add_node(act.code)
# G.add_nodes_from(g_node_data)

for exc in tqdm(ExchangeDataset.select()):
    # print(exc)
    # G.add_edges_from((exc.input_code, exc.output_code, {"weight": exc.data["amount"]}))
    G.add_edge(exc.input_code, exc.output_code)

# Add edges (links)
# edges_data = [
#     ('Node1', 'Node2', {'weight': 3.1415}),
#     # add more edges here
# ]
# G.add_edges_from(edges_data)


def serialize(graph: nx.Graph, path: Path):
    # Save to a pickle file
    pickle.dump(graph, path.open("wb"))


def deserialize(path: Path) -> nx.Graph:
    return pickle.load(path.open("rb"))

# serialize(G, (BASE_DATA_PATH / "graph.nx"))

# does not work for directed graphs
# ordered = topological_sort(G)
# with (BASE_DATA_PATH / "topolical_order.txt").open("w") as fout:
#     for node in ordered:
#         fout.write(node + "\n")



D = dominating_set(G)


import bw2data as bd

bd.projects.set_current("ecoi_dbs")

method = ('CML v4.8 2016 no LT',
          'climate change no LT',
          'global warming potential (GWP100) no LT')

# get main db
db = bd.Database("cutoff391")

# activity: Activity = db.random()
activity = db.get('8cb1567c3ff743d191b309dd64ad84f2')
# 'chromium steel drilling, computer numerical controlled' (kilogram, RoW, None) '8cb1567c3ff743d191b309dd64ad84f2'

# copy some activity to a new db
act_copy = activity.copy("TEST_COPY")
# new_db = bd.Database("new_db")
# new_db.register()
# act_copy["database"] = "new_db"

new_db = bd.Database("new_db")
act_copy = list(new_db)[0]
act_copy_lcia = act_copy.lca(method).score

print(act_copy_lcia)

list(act_copy.exchanges())

for link in activity.exchanges():
    print("IN", link.input, "OUT", link.output, "AMOUNT", link.amount)

for link in activity.upstream():
    print("IN", link.input, "OUT", link.output, "AMOUNT", link.amount)
# ex = myact.new_exchange(input=wind, type="technosphere", amount=1)
# ex.save()




act_copy.save()


#


"""
init of describing characterizations (impact assessment methods) directly from an activity
"""

"""
    "amount": number,  ## This is the only required field
    "uncertainty type": int,
    "loc": number,
    "scale": number,
    "shape": number,
    "minimum": number,
    "maximum": number
    


    0: Undefined or unknown uncertainty.

    1: No uncertainty.

    2: Lognormal distribution. This is a tricky distribution to work with, but is very popular in LCA. The amount field is the median of the data, and the sigma field is the standard deviation of the data when it is log-transformed, i.e. the σ from the formula for the log-normal PDF.

    3: Normal distribution.

    4: Uniform distribution.

    5: Triangular distribution.

    6: Bernoulli distribution.

    7: Discrete uniform.

    8: Weibull.

    9: Gamma.

    10: Beta distribution.

    11: Generalized Extreme Value.

    12: Student’s T.

"""

uncertainty_schema = {
    "type": "object",
    "properties": {
        "amount": {"type": "number"},
        "uncertainty type": {
            "oneOf": [
                {"type": "number", "minimum": 0, "maximum": 12},
                {"type": "string",
                 "enum": ["undefined", "none", "lognormal", "normal", "uniform", "triangular", "bernoulli",
                          "discrete uniform", "weibull", "gamma", "beta", "generalized extreme value", "students t"]},
            ]
        },
        "loc": {"type": "number"},
        "scale": {"type": "number"},
        "shape": {"type": "number"},
        "minimum": {"type": "number"},
        "maximum": {"type": "number"},
    }
}

data = {
    "name": "process 1",
    "code": "process 1",
    "type": "biosphere",
    "characterization": {
        # basic
        ('CML v4.8 2016', 'climate change', 'global warming potential (GWP100)'): 1,
        # uncertainty
        ('CML v4.8 2016', 'climate change', 'global warming potential (GWP100)'): {
            "amount": 1,
            "uncertainty type": 3,
        },
        # location
        ('CML v4.8 2016', 'climate change', 'global warming potential (GWP100)'):
            {"amount": 1,
             "location": "cat"}
        ,
        # checks in an index.
        "global_warming": 1,
    }
}


def read_from_activity(data: dict, metho_index: dict[str, tuple]) -> list[tuple]:
    """
    read characterization data from an activity
    :param data: activity data
    :return: list of characterization data
    """
    characterization = data.get("characterization", {})
    if not isinstance(characterization, dict):
        raise ValueError("Characterization data must be a dictionary")
    # iterate over characterization data
    for key, value in characterization.items():
        # for each key, if its a tuple use that, if its a string, check in some other dict
        if isinstance(key, tuple):
            # check if the tuple is valid
            if not all(isinstance(x, str) for x in key):
                raise ValueError("Characterization keys must be tuples of strings")
        # if string check in some other dict
        elif isinstance(key, str):
            # check if the string is valid
            if key not in metho_index:
                print(f"Warning: {key} is not a valid characterization key")
            # get the tuple from the dict
            key = metho_index[key]

        # check if the value is valid
        if isinstance(value, dict):
            pass



"""
Create/Retrieve sqlite databases from the ecoinvent Excel files (lci, lcia)
The databases are stored in the BASE_DATABASES_PATH folder (data/databases)
Both databases have two tables:
    - activities_{lci|lcia}
        - with the fields: code, name, ... data
            (which is an array with all values, for the exchanges and
            impacts respective order as in the 2nd table)
    - ExchangeInfo | ImpactInfo (names of biosphere flows | impact categories)
        - with the fields:
            ExchangeInfo : ['exchange', 'compartment', 'sub_compartment', 'unit']
            ImpactInfo : ['method', 'category', 'indicator', 'unit']

"""
from enum import Enum
from pathlib import Path
from typing import Type

import openpyxl
from peewee import Model, OperationalError, SqliteDatabase
from playhouse.shortcuts import model_to_dict
from tqdm import tqdm

from enbios2.base.databases import init_databases
from enbios2.base.db_models import (
    EcoinventDatabaseActivity,
    EcoinventResolvedDataset,
    ExchangeInfo,
    ImpactInfo,
    FTS_BW_ActivitySimple,
    BW_Activity,
    EcoinventDataset,
)
from enbios2.const import BASE_DATABASES_PATH
from enbios2.ecoinvent.ecoinvent_index import get_ecoinvent_dataset_index
from enbios2.generic.enbios2_logging import get_logger
from enbios2.generic.util import get_enum_by_value

logger = get_logger(__file__)


class DBTypes(Enum):
    LCI = "lci"
    LCIA = "lcia"
    ActivityFTS = "ActivityFTS"  # experimental


def get_ecoinvent_resolved_entry(name: str) -> EcoinventResolvedDataset:
    db_metadata = EcoinventResolvedDataset.get(EcoinventResolvedDataset.name == name)
    return db_metadata


def set_ecoinvent_resolved_entry(
    db_path: Path, classes: list[Type[Model]]
) -> SqliteDatabase:
    db = SqliteDatabase(db_path)
    for cls in classes:
        cls._meta.database = db
    return db


def get_model_classes(db_type: DBTypes) -> list[Type[Model]]:
    if db_type == DBTypes.LCI:
        return [EcoinventDatabaseActivity, ExchangeInfo]
    if db_type == DBTypes.LCIA:
        return [EcoinventDatabaseActivity, ImpactInfo]
    if db_type == DBTypes.ActivityFTS:
        return [BW_Activity, FTS_BW_ActivitySimple]


def check_exists(name: str) -> bool:
    exists = (
        EcoinventResolvedDataset.select()
        .where(EcoinventResolvedDataset.name == name)
        .exists()
    )
    return exists


def prepare_db(
    db_name: str, db_type: DBTypes, create_tables: bool = True, force_redo: bool = False
) -> tuple[SqliteDatabase, Path, list[Type[Model]]]:
    db_path = BASE_DATABASES_PATH / f"{db_name}.sqlite"
    # call init_ecoinvent_resolved_database
    db = SqliteDatabase(db_path, {"journal_mode": "wal", "cache_size": -1024 * 32})
    db_model_classes = get_model_classes(db_type)
    for model in db_model_classes:
        model._meta.database = db
    # set table name to activity_lci or activity_lcia
    db_model_classes[0]._meta.table_name = f"activity_{db_type.value}"

    if create_tables:
        try:
            db.connect()
        except OperationalError as e:
            logger.error(f"Could not connect to database {db_path}")
            raise e

        if force_redo:
            db.drop_tables(db_model_classes)
        if not db_model_classes[0].table_exists():
            db.create_tables(db_model_classes)

    return db, db_path, db_model_classes


def create(dataset: EcoinventDataset, force_redo: bool = False):
    db_name = "_".join(dataset.identity.split("_")[:-1])
    file_path = dataset.dataset_path
    type_ = get_enum_by_value(DBTypes, dataset.type)
    assert (
        file_path.exists() and file_path.suffix == ".xlsx"
    ), f"File {file_path} does not exist or is not an xlsx file"
    assert type_ in [
        DBTypes.LCI,
        DBTypes.LCIA,
    ], f"Type must be either {DBTypes.LCI} or {DBTypes.LCIA}"
    init_databases()
    db, db_path, db_model_classes = prepare_db(
        db_name,
        type_,
    )

    # count the rows
    activity_count = EcoinventDatabaseActivity.select().count()
    if activity_count == 0:
        reader = openpyxl.load_workbook(
            file_path.as_posix(), read_only=True, data_only=True
        )
        lci_sheet = reader[type_.value.upper()]

        # impacts/exchanges start from G -> 6
        row_generator = lci_sheet.values
        data_index_fields: list[str] = [
            next(row_generator)[6:],
            next(row_generator)[6:],
            next(row_generator)[6:],
            next(row_generator)[6:],
        ]
        index_table = db_model_classes[1]
        data_fields_names = list(index_table._meta.fields.keys())[1:]

        if type_ == DBTypes.LCI:
            assert data_fields_names == [
                "exchange",
                "compartment",
                "sub_compartment",
                "unit",
            ]
        if type_ == DBTypes.LCIA:
            assert data_fields_names == ["method", "category", "indicator", "unit"]

        batch_size = 1000
        batch = []

        def insert_batch(table: Type[Model], batch: list[dict]):
            table.insert_many(batch).execute()
            batch.clear()

        for index, _ in enumerate(data_index_fields[0]):
            batch.append(
                {k: v[index] for k, v in zip(data_fields_names, data_index_fields)}
            )
            if len(batch) == batch_size:
                insert_batch(index_table, batch)

        insert_batch(index_table, batch)

        fields = list(EcoinventDatabaseActivity._meta.fields.keys())[1:]

        for activities in tqdm(row_generator):
            # print({k: v for k, v in zip(fields, activities[:6])})
            db_act = {k: v for k, v in zip(fields, activities[:6])}
            db_act["data"] = activities[6:]
            batch.append(db_act)
            if len(batch) == batch_size:
                insert_batch(EcoinventDatabaseActivity, batch)
        insert_batch(EcoinventDatabaseActivity, batch)

    else:
        print(f"Database '{db_name}' already exists")
    db.close()

    EcoinventResolvedDataset.create(
        name=db_name,
        path=db_path,
        db_type=type_.value,
        ecoinvent_dataset=dataset,
        metadata={"orig_file_name": file_path.name},
    )


def create_from_ecoinvent_dataset(ds: EcoinventDataset):
    if ds.type == "default":  # from valid_ecoinvent_datatypes
        raise ValueError(
            f"Dataset type {ds.type} not supported. Must be either 'lci' or 'lcia'"
        )
    if not ds.xlsx:
        raise ValueError(f"Dataset {ds.name} must be an xlsx file")

    db_name = "_".join(ds.identity.split("_")[:-1])
    if not check_exists(db_name):
        logger.info(f"Creating database '{db_name}'")
        create(ds)
    else:
        logger.info(f"Database '{db_name}' already exists")


def get_activity_data(db_name: str, code: str, drop_zero: bool = False):
    """
    Get the data for an activity in a database
    :param db_name:
    :param code:
    :param drop_zero:
    :return:
    """
    if not check_exists(db_name):
        raise ValueError(f"Database '{db_name}' does not exist")

    db_entry = (
        EcoinventResolvedDataset.select()
        .where(EcoinventResolvedDataset.name == db_name)
        .get()
    )
    db_type = get_enum_by_value(DBTypes, db_entry.db_type)

    db, db_path, db_model_classes = prepare_db(db_name, db_type)

    assert db_type in [
        DBTypes.LCI,
        DBTypes.LCIA,
    ], f"Database type must be either {DBTypes.LCI} or {DBTypes.LCIA}"

    activity = EcoinventDatabaseActivity.get(EcoinventDatabaseActivity.code == code)
    data_index_rows = db_model_classes[1].select()
    data_ = []
    for index, e in enumerate(data_index_rows):
        value = activity.data[index]
        if drop_zero and value == 0:
            continue
        exchange_data = model_to_dict(e, exclude=[db_model_classes[1].id])
        exchange_data["value"] = value
        data_.append(exchange_data)
    return data_


if __name__ == "__main__":
    # EcoinventDataset.select()
    init_databases()
    datasets = get_ecoinvent_dataset_index(type_=["lci", "lcia"], xlsx=True)
    print(list(datasets))
    create_from_ecoinvent_dataset(list(datasets)[1])
    print(
        get_activity_data(
            "cutoff_3.9.1_lci",
            "d195d4a3-6ae5-54e4-9954-7f915cf08668_d69294d7-8d64-4915-a896-9996a014c410",
        )
    )



from dataclasses import asdict

from pydantic.dataclasses import dataclass


class AX:

    @classmethod
    def a(clazz):
        print()
        return clazz


AX.a()
x = AX()
x.a()

@dataclass
class DC:
    a: int

asdict(DC(2))



from peewee import SqliteDatabase

from enbios2.base.db_models import (
    EcoinventResolvedDataset,
    MainDatabase,
    EcoinventDataset,
    BWProjectIndex,
)
from enbios2.const import BASE_DATABASES_PATH


def guarantee_db_dir():
    BASE_DATABASES_PATH.mkdir(parents=True, exist_ok=True)


def init_databases() -> SqliteDatabase:
    guarantee_db_dir()
    database = MainDatabase._meta.database
    database.connect(True)
    if not EcoinventResolvedDataset.table_exists():
        database.create_tables(
            [EcoinventResolvedDataset, EcoinventDataset, BWProjectIndex]
        )
    # EcoinventResolvedDataset.select()
    return database



import numpy as np
from numba import jit
from pint import UnitRegistry

ureg = UnitRegistry()
ureg.define('DALY = []')  # Define DALY as a dimensionless unit
ureg.define('Pt = []')  # Define Pt as a dimensionless unit

from dataclasses import dataclass


# this class is currently not used.
@dataclass
class LCAData:
    techno_matrix: np.ndarray  # Technomatrix
    demand_matrix: np.ndarray  # Demand matrix
    biosphere_matrix: np.ndarray  # Biosphere matrix
    characterization_matrix: np.ndarray  # Characterization matrix
    processes: dict
    demands: dict
    environmental_exchanges: dict
    impact_categories: dict

    def quantities(self):
        # todo: implement
        A_quantities = [ureg.Quantity(1, unit) for unit in A_units]
        np.array([quantity.to_base_units().magnitude for quantity in A_quantities])

        pass


# Technomatrix (A)
A = np.array([[0.1, 0.2, 0.3],
              [0.4, 0.2, 0.1],
              [0.1, 0.3, 0.1]]) * ureg.kg / ureg.kg

processes = {0: {'name': 'Process 1', 'units': 'kg'},
             1: {'name': 'Process 2', 'units': 'kg'},
             2: {'name': 'Process 3', 'units': 'kg'}}

# Demand matrix (F)
F = np.array([[100, 200],
              [300, 400],
              [500, 600]]) * ureg.kg

demands = {0: {'name': 'Demand Scenario 1', 'units': 'units'},
           1: {'name': 'Demand Scenario 2', 'units': 'units'}}

# Biosphere matrix (B)
B = np.array([[0.1, 0.2, 0.3],
              [0.3, 0.2, 0.1]]) * ureg.kg / ureg.kg

environmental_exchanges = {0: {'name': 'Emission 1', 'units': 'kg'},
                           1: {'name': 'Emission 2', 'units': 'kg'}}

# Characterization matrix (C)
C = np.array([[1, 2],
              [3, 4]]) * (ureg.DALY / ureg.kg)

impact_categories = {0: {'name': 'Impact Category 1', 'units': 'DALY/kg'},
                     1: {'name': 'Impact Category 2', 'units': 'Pt/kg'}}


def even_units(A, B):
    pass

# Function to calculate LCIA
def calculate_lcia(lca_data: LCAData):

    # Convert A from Pint Quantity to numpy array, and store the units
    A_units = lca_data.techno_matrix.units
    A = lca_data.techno_matrix.magnitude

    # Calculate the product matrix (I - A)^-1
    I = np.eye(A.shape[0])
    # I = np.eye(A.shape[0]) * ureg.dimensionless

    product_matrix = np.linalg.inv(I - A) * A_units
    # product_matrix = np.linalg.inv(I - A)

    # Calculate the life cycle inventory (LCI) for each demand scenario
    G = product_matrix @ F  # because @ is the matrix multiplication operator for pint arrays
    # G = np.dot(product_matrix, F)
    # Calculate the environmental interventions for each demand scenario
    E = B @ G
    # E = np.dot(B, G)
    # Calculate the life cycle impact assessment (LCIA) for each demand scenario
    H = C @ E
    # H = np.dot(C, E)

    # Individual contributions per process and impact category per scenario
    individual_contributions = []

    for i in range(H.shape[0]):
        impact_category_contributions = []
        for j in range(H.shape[1]):
            process_contributions = G[:, j] * C[i, j]
            impact_category_contributions.append(process_contributions)
        individual_contributions.append(impact_category_contributions)

    return H, individual_contributions


def generate_random_data(num_processes, num_demand_scenarios, num_environmental_exchanges, num_impact_categories):
    # Generate random data for Technomatrix (A)
    techno_matrix = np.random.rand(num_processes, num_processes)

    # Generate random data for Demand matrix (F)
    demand_matrix = np.random.rand(num_processes, num_demand_scenarios)

    # Generate random data for Biosphere matrix (B)
    biosphere_matrix = np.random.rand(num_environmental_exchanges, num_processes)

    # Generate random data for Characterization matrix (C)
    characterization_matrix = np.random.rand(num_impact_categories, num_environmental_exchanges)
    return techno_matrix, demand_matrix, biosphere_matrix, characterization_matrix

@jit(nopython=True, parallel=True)
def numba_generate_random_data(num_processes, num_demand_scenarios, num_environmental_exchanges, num_impact_categories):
    # Generate random data for Technomatrix (A)
    A = np.random.rand(num_processes, num_processes)

    # Generate random data for Demand matrix (F)
    F = np.random.rand(num_processes, num_demand_scenarios)

    # Generate random data for Biosphere matrix (B)
    B = np.random.rand(num_environmental_exchanges, num_processes)

    # Generate random data for Characterization matrix (C)
    C = np.random.rand(num_impact_categories, num_environmental_exchanges)
    return A, F, B, C
    # return A * ureg.kg / ureg.kg, F * ureg.kg, B * ureg.kg / ureg.kg, C * (ureg.DALY / ureg.kg)


def print_random_data(A, F, B, C):
    print("Technomatrix (A):")
    print(A)
    print()

    print("Demand matrix (F):")
    print(F)
    print()

    print("Biosphere matrix (B):")
    print(B)
    print()

    print("Characterization matrix (C):")
    print(C)
    print()


def print_results(H, individual_contributions=None):
    # Print individual contributions per impact category and scenario

    if individual_contributions:
        for i in range(H.shape[0]):
            for j in range(H.shape[1]):
                print(f"Impact Category: {impact_categories[i]['name']}")
                print(f"Demand Scenario: {demands[j]['name']}")
                for k, process_contribution in enumerate(individual_contributions[i][j]):
                    print(f"Process: {processes[k]['name']}, Contribution: {process_contribution}")
                print()

    # Print LCIA results with metadata
    for i in range(H.shape[0]):
        for j in range(H.shape[1]):
            print(f"{impact_categories[i]['name']} for {demands[j]['name']}: {H[i, j]}")


num_processes = 1_000
num_demand_scenarios = 1
num_environmental_exchanges = 40
num_impact_categories = 1


# generate_random_data(num_processes, num_demand_scenarios, num_environmental_exchanges, num_impact_categories)
#
# H, individual_contributions = calculate_lcia(A, F, B, C)
#
# print_results(H, individual_contributions)


def timer(funct, *args, **kwargs):
    import time
    start = time.time()
    res = funct(*args, **kwargs)
    end = time.time()
    # to minutes/seconds
    return end - start, res


def compare_speed():
    # for te,b in ((1_000, 50), (2_000, 50), (3_000, 100), (5_000,200)):
    for te, b in ((5_000, 200), (10_000, 500), (5_000, 200)):
        print("t", te, "b", b)
        t, _ = timer(generate_random_data, te, num_demand_scenarios, b,
                       num_impact_categories)
        print("py", t)
        t, _ = timer(numba_generate_random_data, te, num_demand_scenarios, b,
                       num_impact_categories)
        print("nb", t)


t, data = timer(generate_random_data, te, num_demand_scenarios, b,
               num_impact_categories)


data.quantities()


from numba import njit
import random


@njit(parallel=True)
def numba_monte_carlo_pi(nsamples):
    acc = 0
    for i in range(nsamples):
        x = random.random()
        y = random.random()
        if (x ** 2 + y ** 2) < 1.0:
            acc += 1
    return 4.0 * acc / nsamples


def monte_carlo_pi(nsamples):
    acc = 0
    for i in range(nsamples):
        x = random.random()
        y = random.random()
        if (x ** 2 + y ** 2) < 1.0:
            acc += 1
    return 4.0 * acc / nsamples


def timer(funct, *args, **kwargs):
    import time
    start = time.time()
    funct(*args, **kwargs)
    end = time.time()
    # to minutes/seconds
    return end - start

num = 100_000_000
print(timer(numba_monte_carlo_pi, num))
print(timer(monte_carlo_pi, num))
# 1.5312976837158203
# 32.987468242645264



import pytest

from enbios2.experiment.dot2bw import Dot2BW
import bw2data as bd


@pytest.fixture
def grapgDB_cleanup(request):
    def delete_database():
        bd.projects.set_current("ecoinvent_test")
        bd.Database("graphDB").delete()

    request.addfinalizer(delete_database)


def test_basic(grapgDB_cleanup):
    bd.projects.set_current("ecoinvent_test")

    # bi.bw2setup()
    # bd.Database("graphDB").register()
    # bd.parameters.new_activity_parameters({"some_param": 10}, "graphDB")
    dot2bw = Dot2BW("graphDB", "dot/dot_example.dot")
    print(dot2bw.graph_diff)
    # dot2bw = Dot2BW("graphDB", "dot/dot_example.dot")
    print(dot2bw.graph_diff)
    print("**")
    method = (
        "IPCC 2021",
        "climate change: fossil",
        "global temperature change potential (GTP100)",
    )
    # method_data = bd.Method(method)
    # # print(list(dot2bw.database))
    a = dot2bw.database.get(name="a")
    # #
    print(list(a.biosphere())[0])
    print(a.lca(method=method, amount=1).score)
    # print("+**")
    # ParameterSet
    #
    #
    # parameters = {
    #     'Deep_Thought': {'amount': 42},
    #     'East_River_Creature': {'formula': '2 * Deep_Thought + 16'},
    #     'Elders_of_Krikkit': {'formula': 'sqrt(East_River_Creature)'},
    # }



from copy import copy
from dataclasses import dataclass, field
from pathlib import Path
from typing import Union

import pydot
import bw2data as bd
from bw2data.backends import Activity, Exchange
from bw2data.errors import DuplicateNode

from enbios2.generic.enbios2_logging import get_logger
from enbios2.generic.util import get_data_file_path

logger = get_logger(__file__)


def read_dot_file(file_path: Path) -> pydot.Dot:
    dot_text = file_path.read_text(encoding="utf-8")
    graph, = pydot.graph_from_dot_data(dot_text)
    return graph


@dataclass
class ActivityData:
    name: str
    code: str
    database: str
    location: str
    type: str
    unit: str
    categories: tuple[str]
    exchanges: list[dict[str, any]]


@dataclass
class ExchangeData:
    amount: float
    type: str


def collect_graph_node_names(graph: pydot.Dot) -> set[str]:
    edges = graph.get_edges()
    all_nodes = set()  # the implicit ones in the edges

    for edge in edges:
        source = edge.get_source()
        target = edge.get_destination()
        all_nodes.update([source, target])

    return all_nodes


def node2code(node: pydot.Node):
    return f"_{node.get_name()}_"


def activity2node(activity: Activity):
    return activity["code"][1:-1]


def parse_node_attributes(node: pydot.Node):
    attributes: dict[str, str] = copy(node.get_attributes())
    for k, v in attributes.items():

        if v.startswith('"') and v.endswith('"'):
            v = v[1:-1]
        # split lists # todo assumes there are no " in the list with commas in the strings.
        if v.startswith("[") and v.endswith("]"):
            v = v[1:-1]
            v = v.split(",")
            v = [part.strip() for part in v]
        attributes[k] = v
    return attributes


def get_external_activities(graph: pydot.Dot):
    external_activities_nodes = graph.get_nodes()
    external_activity_map: dict[str, Activity] = {}

    for node in external_activities_nodes:
        attributes = parse_node_attributes(node)
        if "database" in attributes:
            # more attributes, like code, location
            activity = bd.Database(attributes["database"]).get(name=attributes["name_"],
                                                               categories=tuple(attributes["categories"]),
                                                               type=attributes["type"])
            logger.debug("external activity : %s", attributes["name_"])
            external_activity_map[node.get_name()] = activity
        else:
            pass
            # no database attribute, so it is a reference to an activity in the current database
    return external_activity_map


@dataclass
class GraphData:
    activities: list[Activity]
    edges: list[ExchangeData]


@dataclass
class GraphDiffData:
    removed_activities: list[Activity] = field(default_factory=list)
    added_activities: list[Activity] = field(default_factory=list)
    removed_edges: list[Exchange] = field(default_factory=list)
    added_edges: list[Exchange] = field(default_factory=list)


class Dot2BW:

    def __init__(self, database_name: str = "graphDB", dot_file_path: Union[str, Path] = None):
        self.database_name = database_name
        if self.database_name not in bd.databases:
            bd.Database(database_name).register()
        self.database = bd.Database(database_name)

        self.graph_diff = GraphDiffData()

        if dot_file_path:
            dot_file_path = get_data_file_path(dot_file_path)
            graph = read_dot_file(dot_file_path)
            self.define_db_with_graph(graph)

    @staticmethod
    def edge_sign(edge: pydot.Edge) -> str:
        return f"_{edge.get_source()}>{edge.get_destination()}_"

    @staticmethod
    def resolve_edge_sign(edge_sign: str) -> tuple[str, str]:
        from_act, to_act, _ = edge_sign[1:-1].split(">")
        return from_act, to_act

    def define_db_with_graph(self, graph: pydot.Dot) -> GraphDiffData:
        """
        Define a database with the activities based on the dot graph.
        Returns new newly defined activities and edges.
        :param graph:
        :return:
        """

        # get all activities in the graphDB
        internal_act_map = {activity2node(act): act for act in self.database}
        # logger.debug(act_map.keys())

        # get all implicit nodes in the graph
        node_names: set[str] = collect_graph_node_names(graph)
        # print("node_names", node_names)
        activities_to_remove = set(internal_act_map.keys()) - node_names
        # print("activities_to_remove", activities_to_remove)
        # delete all activities that are not in the graph
        for act in activities_to_remove:
            internal_act_map[act].delete()
            self.graph_diff.removed_activities.append(internal_act_map[act])

        # get all external activities in the graph (defines explicitly in the graph)
        all_act_map = copy(internal_act_map)
        all_act_map.update(get_external_activities(graph))
        activities_to_add = node_names - set(all_act_map.keys())
        # print("activities_to_add", activities_to_add)

        for node in activities_to_add:
            activity = self.add_activity(node)
            internal_act_map[node] = activity
            all_act_map[node] = activity
            self.graph_diff.added_activities.append(activity)

        node_out_edges = {}
        for edge in graph.get_edges():
            node_out_edges.setdefault(edge.get_source(), []).append((edge.get_destination(), edge))

        for node_name, activity in internal_act_map.items():
            node_edges = node_out_edges.get(node_name, [])

            destination_nodes, edges = zip(*node_edges)

            exchanges = list(activity.exchanges())

            # print(node_edges, exchanges)
            # delete all exchanges that are not defined in the graph
            for ex in exchanges:
                if self.resolve_edge_sign(ex.get("sign"))[1] not in destination_nodes:
                    ex.delete()
                    self.graph_diff.removed_edges.append(ex)
            # add all edges that are not defined in the database
            edge_signs = [ex.get("sign") for ex in exchanges]
            for edge in edges:
                if self.edge_sign(edge) not in edge_signs:
                    from_act: Activity = all_act_map.get(edge.get_source())
                    to_act: Activity = all_act_map.get(edge.get_destination())
                    # print(from_act, to_act)
                    attributes = edge.get_attributes()
                    amount = attributes.get("amount", 1)
                    type_ = attributes.get("type", "technology")
                    unit = attributes.get("unit")

                    # print(from_act, to_act)
                    sign = self.edge_sign(edge)

                    print(f"adding exchange: {from_act['name']}, {to_act['name']}, {type_}, {amount}, {unit}, {sign}")
                    ex:Exchange = from_act.new_exchange(input=to_act, type=type_, amount=amount, unit=unit,
                                               sign=sign)
                    ex.save()
                    self.graph_diff.added_edges.append(ex)

        return self.graph_diff

    def add_activity(self, name: str) -> Activity:
        try:
            act = self.database.new_activity(name=name, code=f"_{name}_")
            act.save()
            print(f"activity saved: {name}")
            return act
        except DuplicateNode:
            logger.debug(f"Duplicate node not added {name}")
            pass



from pint import Quantity


def compact_all_to(quantities: list[Quantity], use_min: bool = True) -> list[Quantity]:
    """
    Convert all quantities to the same unit, and return the compacted values
    :param quantities:
    :param use_min: use the unit of the smallest quantity (otherwise use the largest)
    :return: a list of compacted quantities with the same unit
    """
    base_func = min if use_min else max
    base_value = base_func([q.to_compact() for q in quantities])
    return [q.to(base_value.units) for q in quantities]



from enbios2.base.unit_registry import ureg


# ureg.define('Wh_from_battery = 3 * kg')  # Define a new unit
# # ureg.define('kWh_from_batteries = 3 * kg / kWh')  # Define a new unit
#
# target_quantity = ureg.parse_expression("Wh_from_batteries", case_sensitive=False) * 1
# conv = target_quantity.to("kilogram")
# print(conv)
# bw2data.projects.set_current("ecoinvent")
# db = bw2data.Database("cutoff_3.9.1_default")
# battery = db.get_node("a2634d3b6d27046ccb1d0af9c0384c64")


# conv = target_quantity.to(battery["unit"])

# print(conv)

# target_quantity_Wh = ureg.Quantity(1, 'Wh_from_batteries')
# conv_Wh = target_quantity_Wh.to("kg")
# print(f'1 Wh_from_batteries is {conv_Wh.magnitude} kg')

# target_quantity_kWh = ureg.Quantity(1, 'kWh_from_batteries')
# conv_kWh = target_quantity_kWh.to("kg")
# print(f'1 kWh_from_batteries is {conv_kWh.magnitude} kg')


# wh_w = ureg("kilohour * watt")
# kw_x_h = ureg("kilowatt * hour")
# kw_x_h.to_preferred()
# kw_h = ureg("kilowatt_hour")

us = [
    ureg("kilowatt_hour") * 1,
    ureg("kilowatt_hour") * 2,
    ureg("kilowatt_hour") * 3,
    ureg("kilowatt_hour") * 40000,
    ureg("megawatt_hour") * 3,
]
c = us[3].to_compact()



# try full text search like
# https://charlesleifer.com/blog/using-sqlite-full-text-search-with-python/
from pathlib import Path
from typing import Optional

import bw2data
from bw2data.backends import ActivityDataset
from playhouse.sqlite_ext import *
from tqdm import tqdm

from enbios2.ecoinvent.spatial import geo_code2name, get_ecoinvent_geo_data
from enbios2.experiment.databases import DBTypes, create_db
from enbios2.experiment.db_models import BW_Activity, FTS_BW_ActivitySimple
from enbios2.generic.enbios2_logging import get_logger
from enbios2.generic.files import ReadDataPath

logger = get_logger(__file__)

def build_fts(bw_project: str, db_path: Path, ecoinvent_geo_xml_path: Optional[str] = None):
    logger.info(f"Building FTS for {bw_project} to {db_path}")
    create_db(db_path, db_type=DBTypes.ActivityFTS)
    bw2data.projects.set_current(bw_project)
    activities = ActivityDataset.select()

    # content_fields = ["name", "product", "comment"]

    geo_code2_name = {}
    if ecoinvent_geo_xml_path:
        geo_code2_name = geo_code2name(get_ecoinvent_geo_data(ReadDataPath(ecoinvent_geo_xml_path)))

    for act in tqdm(activities):
        db_a = BW_Activity.create(code=act.code, database=act.database, name=act.name, location=act.location,
                                  location_name=geo_code2_name.get(act.location, ""),
                                  product=act.product if act.product else "",
                                  type=act.type, comment=act.data.get("comment", ""),
                                  synonyms="##".join(act.data.get("synonyms", "")),
                                  classification=act.data.get("classification", ""),
                                  unit=act.data.get("unit", ""),
                                  reference_product=act.data.get("reference_product", ""))

        # co = (getattr(db_a, field, "") for field in content_fields)
        # co_s = "\n".join(f for f in co if f)
        FTS_BW_ActivitySimple.create(docid=db_a.id,
                                     name=db_a.name,
                                     comment=db_a.comment,
                                     location_name=db_a.location_name,
                                     product=db_a.product,
                                     synonyms=db_a.synonyms)


def _parse_match_info(buf):
    bufsize = len(buf)  # Length in bytes.
    return [struct.unpack('@I', buf[i:i + 4])[0] for i in range(0, bufsize, 4)]


def rank(raw_match_info):
    # handle match_info called w/default args 'pcx' - based on the example rank
    # function http://sqlite.org/fts3.html#appendix_a
    match_info = _parse_match_info(raw_match_info)
    score = 0.0
    p, c = match_info[:2]
    for phrase_num in range(p):
        phrase_info_idx = 2 + (phrase_num * c * 3)
        for col_num in range(c):
            col_idx = phrase_info_idx + (col_num * 3)
            x1, x2 = match_info[col_idx:col_idx + 2]
            if x1 > 0:
                score += float(x1) / x2
    return -score


def search(text: str):
    query = (BW_Activity
             .select(BW_Activity.name)
             .join(FTS_BW_ActivitySimple, on=(BW_Activity.id == FTS_BW_ActivitySimple.docid))
             .where(FTS_BW_ActivitySimple.match(text)).order_by(FTS_BW_ActivitySimple.rank()).limit(20)
             .dicts())
    for row_dict in query:
        print(row_dict)


build_fts("ecoi_dbs",
          Path('fts_db/cutoff391.db'),
          "ecoinvent/ecoinvent 3.9.1_cutoff_ecoSpold02/MasterData/Geographies.xml")



import contextlib
import io
import re
from enum import Enum
from pathlib import Path
from time import time
from typing import Union, Type, Any

from enbios2.const import BASE_DATA_PATH
from enbios2.generic.enbios2_logging import get_logger

logger = get_logger(__file__)


def generate_levensthein_name_map(
    names_a: list[str], names_b: list[str]
) -> dict[str, str]:
    try:
        from Levenshtein import ratio
    except ImportError:
        raise ImportError(
            "Levensthein module not found. Install with `pip install Levensthein`"
        )

    names_map: dict[str, str] = {}
    remaning_names = names_b.copy()
    for term in names_a:
        if term not in remaning_names:
            # find closest match
            closest_match = max(remaning_names, key=lambda x: ratio(term, x))
            names_map[term] = closest_match
            remaning_names.remove(closest_match)
    return names_map


def generate_levensthein_dict_map(
    names_a: list[str], dicts: list[dict], dict_key: str
) -> dict[str, dict]:
    try:
        from Levenshtein import ratio
    except ImportError:
        raise ImportError(
            "Levensthein module not found. Install with `pip install Levensthein`"
        )

    names_map: dict[str, dict] = {}
    remaning_dicts = dicts.copy()
    for term in names_a:
        # find closest match
        closest_match = max(remaning_dicts, key=lambda x: ratio(term, x[dict_key]))
        names_map[term] = closest_match
        remaning_dicts.remove(closest_match)
    return names_map


def get_data_file_path(path: Union[str, Path]) -> Path:
    """
    Returns a Path object from the data directory
    :param path:
    :return:
    """
    path = Path(path)
    if path.is_absolute():
        return path
    else:
        return BASE_DATA_PATH / path


def safe_name(text: str, replacer_char: str = "_") -> str:
    """
    Replace characters that are not allowed in file names with a replacer character.
    :param text: text to convert
    :param replacer_char: character to replace with
    :return: sane file name
    """
    if len(replacer_char) > 1:
        logger.warning("replacer_char should be a single character")
        replacer_char = replacer_char[0]
    return re.sub(r"[/\\?%*:|\"<>\x7F\x00-\x1F]", replacer_char, text)


def simple_name(text, replacer_char: str = "_") -> str:
    """
    Replace characters that are not  alphanumeric of _ with a replacer character.
    :param text:
    :param replacer_char: should be just a single character
    :return:
    """
    if len(replacer_char) > 1:
        logger.warning("replacer_char should be a single character")
        replacer_char = replacer_char[0]
    return re.sub(f"[^0-9a-zA-Z{replacer_char}]", replacer_char, text)


def get_enum_by_value(enum_type: Type[Enum], value):
    for enum_member in enum_type.__members__.values():
        if enum_member.value == value:
            return enum_member
    return None


def call_muted(func, *args, **kwargs) -> Any:
    """
    Call a function
    :param func: function to call
    :param args: function args
    :param kwargs: function kwargs
    :return: function return value
    """
    with contextlib.redirect_stdout(io.StringIO()):
        return func(*args, **kwargs)


@contextlib.contextmanager
def muted():
    with contextlib.redirect_stdout(io.StringIO()):
        yield


@contextlib.contextmanager
def timed():
    start = time()
    try:
        yield
    finally:
        print(f"({time() - start:.2f}s)")
